from cbr_website_beta.flask.Flask_Site import Flask_Site
from cbr_website_beta.flask.filters.Current_User import client__logged_in

class API_Request:
    def __init__(self, path='/', username=None, user_groups=None, app=None):
        self.path         = path
        self.app          = app or Flask_Site().app()
        if username:
            self.client = client__logged_in(app=self.app, user_name=username, user_groups=user_groups)
        else:
            self.client   = self.app.test_client()
        self.response     = None


    def __enter__(self):
        self.response    = self.client.get(self.path)
        return self, self.response.json

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


from flask import request

from cbr_website_beta           import static
from osbot_utils.utils.Files    import path_combine

FLASK_APP__STATIC_FOLDER    = 'dist'
FLASK_APP__STATIC_PATH      = path_combine(static.path, FLASK_APP__STATIC_FOLDER)
FLASK_APP__STATIC_URL       = f'/{FLASK_APP__STATIC_FOLDER}'
FLASK_APP__STATIC_RULE_NAME = f'/{FLASK_APP__STATIC_FOLDER}/<path:filename>'


class Flask_Utils:
    @staticmethod
    def is_static_request():
        return request.path.startswith(FLASK_APP__STATIC_URL)



# from osbot_aws.aws.dynamo_db.DyDB__Timeseries import DyDB__Timeseries
# from osbot_utils.utils.Misc import timestamp_utc_now
#
#
# class DyDB__Chat_Threads(DyDB__Timeseries):
#
#     def __init__(self, **kwargs):
#         super().__init__(**kwargs)
#         self.table_name          = 'tcb_chat_threads'
#         self.key_name            = 'thread_id'
#         self.key_attribute_type  = 'S'
#         self.index_name          = 'chats_by_user'
#         self.partition_key_name  = 'user'
#         self.partition_key_value = 'user_x'
#         self.data_field_name     = 'chat_thread'
#
#     def add_document(self, document, partition=None):
#         if self.key_name not in document:
#             document[self.key_name] = self.dynamo_db.random_id()     # If key is present, generate a random UUID as the key
#         return self.dynamo_db.document_add(table_name=self.table_name, key_name=self.key_name, document=document)
#
#     def add_chat_thread(self, user_name, thread_id, index, data_type, data):
#         data['timestamp'] = timestamp_utc_now()                         # make sure each request has a unique timestamp
#         data['index'    ] = index                                       # for now store the index in the data
#         data['data_type'] = data_type                                   # for now store the data_type in the data
#         document = { self.primary_key         : self.dynamo_db.random_id()  ,
#                      self.key_name            : thread_id              ,
#                      self.partition_key_name  : user_name                   ,
#                      self.data_field_name     : data                        }
#
#         return self.add_document(document)
#
#     def documents(self,partition=None):
#         return self.dynamo_db.documents_all(table_name=self.table_name)


# -*- encoding: utf-8 -*-
"""
Copyright (c) 2019 - present AppSeed.us
"""

import os
from cbr_website_beta.apps.CBR_Config   import cbr_config
from cbr_website_beta.utils.Site_Utils  import Site_Utils
from cbr_website_beta.utils.Version import version
from cbr_website_beta.utils.Web_Utils   import Web_Utils



class Config(object):
    ENV         = cbr_config.env()
    ASSETS_ROOT = os.getenv('ASSETS_ROOT', cbr_config.assets_root())
    VERSION     = version
    ASSETS_DIST = cbr_config.assets_dist()

    
class ProductionConfig(Config):
    DEBUG = False

    # Security
    SESSION_COOKIE_HTTPONLY  = cbr_config.session_cookie_httponly  ()
    REMEMBER_COOKIE_HTTPONLY = cbr_config.remember_cookie_httponly ()
    REMEMBER_COOKIE_DURATION = cbr_config.remember_cookie_duration ()
    LOGIN_ENABLED            = cbr_config.login_enabled            ()

class DebugConfig(Config):
    DEBUG = True


# Load all possible configurations
config_dict = {
    'Production': ProductionConfig,
    'Debug'     : DebugConfig
}



import traceback

from cbr_website_beta._cbr_shared.schemas.CBR_Logging import CBR_Logging
from cbr_website_beta.apps.CBR_Config import cbr_config
from cbr_website_beta.utils.Web_Utils import Web_Utils
from osbot_aws.aws.dynamo_db.domains.DyDB__Table_With_Timestamp import DyDB__Table_With_Timestamp
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Misc import date_time_now

DYNAMO_DB__TABLE_NAME__CBR_LOGGING = '{env}__cbr_logging'
#TABLE_CBR_LOGGING__INDEXES_NAMES   = [ 'date', 'level', 'message', 'source', 'topic']

class DyDB__CBR_Logging(DyDB__Table_With_Timestamp):

    env: str = cbr_config.env()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.table_name    = DYNAMO_DB__TABLE_NAME__CBR_LOGGING.format(env=self.env)
        self.disabled       = cbr_config.aws_enabled()
        #self.table_indexes = TABLE_CBR_LOGGING__INDEXES_NAMES

    def date_today(self):
        return date_time_now(date_time_format='%Y-%m-%d')

    def documents__today(self):
        if self.disabled:
            return []
        index_name  = 'date'
        index_type  = 'S'
        index_value = date_time_now(date_time_format='%Y-%m-%d')
        documents = self.query_index(index_name=index_name, index_type=index_type, index_value=index_value)
        return documents

    def log_error(self, message, log_data):
        if self.disabled:
            return
        try:
            with CBR_Logging() as _:
                _.date        = self.date_today()
                _.message     = message
                _.level       = 'ERROR'
                _.source      = 'odin'
                _.status_code = '500'
                _.extra_data  = log_data
                result = self.add_document(_.json())
                document_id = result.get('document', {}).get('id')
                print(f'>>>> added error as {document_id} to {self.table_name}')

        except Exception as e:
            print('>>>>> MAJOR ERROR in log_error <<<<')
            pprint(e)               # todo: add to logging

    def log_exception(self, exception):
        if self.disabled:
            return
        stack_trace = traceback.format_exc()
        log_data = {
            'error': str(exception),
            'stack_trace': stack_trace,
            'description': 'Internal Server Error occurred.'
        }
        message = f"ERROR 500: { log_data.get('error') }"
        self.log_error(message, log_data)

    def add_log_message(self, message, **kwargs):
        if self.disabled:
            return
        cbr_logging = CBR_Logging(message=message, **kwargs)
        cbr_logging.date = self.date_today()
        document = cbr_logging.json()
        result = self.add_document(document)
        document_id = result.get('document', {}).get('id')
        return document_id

    def query__today__last_n_hours(self, hours=1):
        if self.disabled:
            return []
        filter_key   = 'date'
        filter_value = self.date_today()
        documents = self.query_index_last_n_hours(filter_key, filter_value, hours)
        return documents

    def query__today__last_n_hours__where_header__matches(self, header_name, header_value, hours=1):
        if self.disabled:
            return []
        filter_key   = 'date'
        filter_value = self.date_today()
        query_filter = { 'filter_expression'     : "#extra_data.#event.#headers.#websocket_key = :websocket_key_value",
                         'expression_attr_names' : { '#extra_data'         : 'extra_data'        ,
                                                     '#event'              : 'event'             ,
                                                     '#headers'            : 'headers'           ,
                                                     '#websocket_key'      : header_name},
                         'expression_attr_values': { ':websocket_key_value': {'S': header_value} }}
        documents = self.query_index_last_n_hours(filter_key, filter_value, hours, query_filter)
        return documents

    def query__today__last_n_hours__where_field__contains(self, field_name, field_value, hours=1, field_type='S'):
        if self.disabled:
            return []
        filter_key      = 'date'
        filter_value    = self.date_today()
        query_filter    = {'filter_expression'     : "contains(#key, :value)"              ,
                           'expression_attr_names' : { '#key'  : field_name               },
                           'expression_attr_values': { ':value': {field_type: field_value}}}
        documents = self.query_index_last_n_hours(filter_key, filter_value, hours, query_filter)
        return documents


dydb_cbr_logging = DyDB__CBR_Logging()


# todo: remove code after checking that we have moved all into athena repo
# from logging import INFO
# from os import environ
#
# from dotenv import load_dotenv
#
# from osbot_aws.apis.EC2 import EC2
# from osbot_aws.helpers.EC_Instance import EC2_Instance
# from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
# from osbot_utils.decorators.methods.cache_on_self import cache_on_self
# from osbot_utils.testing.Logging import Logging
# from osbot_utils.utils.Files import file_name_without_extension
# from osbot_utils.utils.Http import wait_for_ssh
#
# CBR__AMIS = {'test-image-creation':'ami-0136026a91d5f4151'}
#
# class CBR__EC2_Instance(EC2_Instance):
#     logging : Logging
#
#     def __init__(self):
#         super().__init__()
#         self.create_kwargs = self.cbr_create_kwargs()
#         self.logging.log_to_sys_stdout()
#         self.logging.set_logger_level(INFO)
#
#     def cbr_create_kwargs(self):
#         load_dotenv()
#         instance_type     = 't3.nano'
#         spot_instance     = True
#         security_group_id = environ.get('EC2_TESTS__SECURITY_GROUP_ID'     )
#         ssh_key_name      = environ.get('EC2_TESTS__PATH_SSH_KEY_FILE_NAME')
#         return  dict(image_id          = 'ami-0136026a91d5f4151', # AMI created with python installed
#                      key_name          = ssh_key_name  ,
#                      security_group_id = security_group_id,
#                      instance_type     = instance_type      ,
#                      spot_instance     = spot_instance      )
#     def log(self, message):
#         self.logging.info(message)
#
#     def ssh(self):
#         ssh_key_file = environ.get('EC2_TESTS__PATH_SSH_KEY_FILE')
#         ssh_key_user = environ.get('EC2_TESTS__PATH_SSH_KEY_USER')
#         return super().ssh(ssh_key_file=ssh_key_file, ssh_key_user=ssh_key_user)
#
#
#     def wait_for_instance_running(self):
#         self.ec2.wait_for_instance_running(self.instance_id)
#
#     def wait_for_ssh(self):
#         return wait_for_ssh(self.ip_address())
#
#
#
#
#
#



# import boto3
# from osbot_aws.helpers.IAM_User import IAM_User
#
# from osbot_aws.helpers.IAM_Role import IAM_Role
#
# from osbot_aws.apis.IAM import IAM
#
#
# class Logs_Permission_Checker:
#
#     def __init(self):
#         pass
#
#     def get_current_role_name(self):
#         sts = boto3.client('sts')
#         identity = sts.get_caller_identity()
#         arn_parts = identity['Arn'].split(':')
#         role_name = arn_parts[-1].split('/')[-1]
#         return role_name
#
#     def get_role_policies(self, role_name):
#         iam = boto3.client('iam')
#         policies = iam.list_attached_role_policies(RoleName=role_name)
#         return [policy['PolicyArn'] for policy in policies['AttachedPolicies']]
#
#     def get_policy_permissions(self, policy_arn):
#         iam = boto3.client('iam')
#         policy = iam.get_policy_version(
#             PolicyArn=policy_arn,
#             VersionId='v1'  # You might want to find the correct version dynamically
#         )
#         return policy['PolicyVersion']['Document']['Statement']
#
#
#     def check_permissions(self, required_permissions):
#         role_name = self.get_current_role_name()
#         print('----------: role_name', role_name)
#
#         iam_user = IAM_User(user_name=role_name)
#         print(f'----iam_user: policies     : {iam_user.policies()}'     )
#         print(f'----iam_user: user policies: {iam_user.user_policies()}')
#
#         iam_role = IAM_Role(role_name=role_name)
#         print(f'----iam_role: policies     : { list(iam_role.policies())}')
#         print(f'----iam_role: policies_statements: {iam_role.policies_statements()}')
#         policy_arns = iam_user.user_policies()
#         return policy_arns
#
#         # role_name = self.get_current_role_name()
#         # policy_arns = self.get_role_policies(role_name)
#
#         for policy_arn in policy_arns:
#             permissions = self.get_policy_permissions(policy_arn)
#             for permission in permissions:
#                 action = permission['Action']
#                 if set(action).intersection(set(required_permissions)):
#                     print(f"Required permissions are present in policy: {policy_arn}")
#                     return True
#         print("Required permissions are not present.")
#         return False



from flask import render_template, g
from cbr_website_beta.flask.filters.Current_User import Current_User


def view__tree_view():
    g.trace_call.trace_call_handler.stack.add_node(title='-- in view__tree_view')
    with g.trace_call:
        current_user = Current_User()
        current_user.user_data_from_s3()

    # view_model = g.trace_call.view_model
    # g.trace_call.print_traces()
    return render_template('dev/tree_view.html', title='Tree View')


from cbr_website_beta.flask.processors.date_time_now import date_time_now
from cbr_website_beta.flask.processors.menu_links import menu_links


def register_processors(app):

    @app.context_processor
    def inject__date_time_now():
        return {'date_time_now': date_time_now()}

    @app.context_processor
    def inject__menu_links():
        return { 'menu_links' : menu_links }


from cbr_website_beta.flask.hooks.add_cookies_to_g      import add_cookies_to_g
from cbr_website_beta.flask.hooks.allow_anonymous       import allow_anonymous, admins_only
from cbr_website_beta.flask.hooks.populate_variable_g   import populate_variable_g


def register_hooks(app):
    app.before_request(add_cookies_to_g     )
    app.before_request(populate_variable_g  )
    app.before_request(admins_only          )
    app.before_request(allow_anonymous      )
    #app.after_request (set_test_cookies     )


# from cbr_website_beta.bots.Minerva_Rest_API import Minerva_Rest_API
#
# SERVICES_TO_SKIP = ['Amazon Registrar', 'Tax', 'Refund', 'AWS Key Management Service', 'Amazon Relational Database Service',
#                     'AWS CloudTrail','AmazonCloudWatch','AWS Secrets Manager',
#                     'Amazon Virtual Private Cloud','Amazon Virtual Private Cloud',
#                     'Amazon Simple Notification Service','AWS Step Functions',
#                     'Amazon Route 53',
#                     #'Amazon API Gateway', 'Amazon EC2 Container Registry (ECR)'
#                     ]
#
# class Chart_Data__AWS__Minerva:
#
#     def __init__(self):
#         self.bot_minerva       = Minerva_Rest_API()
#
#     def aws_cost_explorer(self):
#         cost_data    = self.bot_minerva.aws_cost_explorer()
#         indexed_data = {}
#
#         for day, services in cost_data.items():
#             for service, info in services.items():
#                 if service in SERVICES_TO_SKIP:
#                     continue
#                 if service not in indexed_data:
#                     indexed_data[service] = {}
#
#                 indexed_data[service][day] = {'cost': info['cost']}
#
#         all_services = list(indexed_data.keys())
#         all_days    = list(cost_data.keys())
#
#         all_costs = []
#         for service in all_services:
#             day_data = []
#             for day in all_days:
#                 if day not in indexed_data.get(service):
#                     value = 0
#                 else:
#                     value = indexed_data[service][day]['cost']
#                 day_data.append(value)
#             all_costs.append(day_data)
#
#
#         chart_data= dict(all_services=all_services,
#                          all_days=all_days,
#                          all_costs=all_costs)
#         return chart_data


from flask import current_app

from osbot_utils.utils.Dev import pprint


def menu_links(prefix):
    url_map = current_app.url_map                           # Get the current Flask app's URL map
    matching_routes = {}                                    # Initialize an empty dictionary for matching routes

    rules = {str(rule):rule for rule in url_map.iter_rules()}

    for rule_str, rule in sorted(rules.items()):            # Loop through each rule in the  (sorted) URL map
        if '/api' in rule_str:                              # Skip any API routes
            continue
        if '<' in rule_str:                                 # Skip any routes with dynamic parts
            continue
        if 'POST' in rule.methods:                          # Skip any POST routes
            continue
        if rule_str.startswith(prefix):                     # Check if the rule starts with the prefix
            last_part = rule_str.split('/')[-1]             # Extract the last part of the route
            title = last_part.replace('-', ' ').title()     # Replace hyphens with spaces and title-case it

            if title == '':
                title = prefix[1:].title()                  # If it's the root route, use the prefix itself (without slash)

            matching_routes[title] = rule_str               # Add to dictionary with title as key and rule as value
    return matching_routes                                  # Return the dictionary of matching routes


# import json
#
# from cbr_website_beta._cbr_shared.dynamo_db.DyDB__CBR_Logging       import dydb_cbr_logging
#
# from cbr_website_beta.aws.apigateway.web_sockets.WS__Handle_Lambda  import WS__Handle_Lambda
# from cbr_website_beta.utils.Site_Utils                              import Site_Utils
# from osbot_utils.utils.Misc                                         import wait_for
#
#
# class WS__Handle_Events__To_Logs(WS__Handle_Lambda):
#     site_utils : Site_Utils
#     version    : str
#     def __init__(self):
#         super().__init__()
#         self.version = self.site_utils.version()
#
#     def route__connect(self, event, connection_id):
#         self.send_message(f'Connect requested', connection_id, event)
#         return super().route__connect(event, connection_id)
#
#     def route_disconnect(self, event, connection_id):
#         self.send_message(f'Disconnect requested', connection_id, event)
#         return super().route_disconnect(event, connection_id)
#
#     def route_default(self, event, connection_id):
#         body = event.get('body', 'NA')
#         self.send_message     (f'Default route reached with connection ID', connection_id, event)
#         self.sent_message_back('on DEFAULT', connection_id, body)
#         return super().route_default(event, connection_id)
#
#     def route_unknown(self, event, connection_id, route_key):
#         self.send_message(f'Unknown routeKey:' + route_key, connection_id, event)
#         return super().route_unknown(event, connection_id, route_key)
#
#     def send_message(self, route_key, ws_connection_id, event):
#         message = f'[{self.version}] ws message "{route_key}" with connection id: {ws_connection_id}'
#         kwargs = dict(message    = message        ,
#                       extra_data = {'event': event})
#         dydb_cbr_logging.add_log_message(**kwargs)
#
#
#     def sent_message_back(self, event, connection_id, body):
#         try:
#             apiId = 'ow2twc6ee4'
#             region = 'eu-west-2'
#             stage = 'prod'
#             endpoint_url = f'https://{apiId}.execute-api.{region}.amazonaws.com/{stage}'
#
#             import boto3
#             client = boto3.client('apigatewaymanagementapi',
#                                   endpoint_url=endpoint_url)
#             wait_for(2)
#             reply_data = {'message': 'CCCCC Hi: ', 'event': event, "body": body}
#             client.post_to_connection(ConnectionId=connection_id, Data=json.dumps(reply_data))
#             kwargs = dict(message=f'sent message back to client : {reply_data}', extra_data=reply_data)
#             dydb_cbr_logging.add_log_message(**kwargs)
#         except Exception as e:
#             kwargs = dict(message=f'error in sent_message_back: {e}')
#             dydb_cbr_logging.add_log_message(**kwargs)


from functools import cache

from cbr_website_beta.aws.s3.S3_DB_Base import S3_DB_Base
from osbot_utils.utils.Misc             import lower, random_string_and_numbers
from cbr_website_beta.aws.s3.DB_User    import DB_User

class DB_Users(S3_DB_Base):

    def db_user(self, user_id):
        return DB_User(user_id)

    def db_users(self):
        db_users = []
        for user_id in self.db_users_ids():
            db_users.append(self.db_user(user_id))
        return db_users

    def db_users_ids(self):
        # todo: refactor all these .s3() calls to helper methods
        # todo: refactor this method to use self.s3().folder_list(s3_bucket=self.s3_bucket(), parent_folder=parent_folder, return_full_path=return_full_path)
        s3_keys = self.s3().find_files(self.s3_bucket(), self.s3_folder_users_metadata())  # Fetch S3 keys for user metadata files
        usernames = []                                                                           # Initialize an empty list to store usernames
        for s3_key in s3_keys:
            filename = s3_key.split('/')[-1]                                                     # Extract the filename from the S3 key
            username = filename.replace('.json', '')                                             # Remove the '.json' extension to get the username
            usernames.append(username)                                                           # Append the username to the list

        return usernames

    def random_db_user(self, prefix='pytest_user'):
        user_id = self.random_user_id(prefix=prefix)
        return self.db_user(user_id)

    def random_user_id(self, prefix='pytest_user'):
        return f'{prefix}__{lower(random_string_and_numbers())}'

    def users__all_metadatas(self,max=None):
        user_data = {}
        db_users = self.db_users()
        if max:
            db_users = db_users[:max]
        for db_user in db_users:
            user_data[db_user.user_id] = db_user.metadata()

        return user_data



from osbot_aws.aws.ecs.ECS import ECS
from osbot_aws.aws.ecs.ECS_Fargate_Task import ECS_Fargate_Task
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.decorators.methods.cache_on_self import cache_on_self


class CBR__Fargate(Kwargs_To_Self):
    cluster_name : str
    image_name   : str

    @cache_on_self
    def ecs(self):
        return ECS()

    def cluster(self):
        cluster_arn = self.ecs().cluster_arn(self.cluster_name)
        return self.ecs().cluster(cluster_arn=cluster_arn)

    def task(self):
        task_definition_arn = 'arn:aws:ecs:eu-west-2:470426667096:task-definition/task-definition-video-editing'
        task                = ECS_Fargate_Task(cluster_name=self.cluster_name, image_name=self.image_name)
        task.task_family    = task_definition_arn
        return task

    def task_logs(self, task_arn, task_definition_arn):
        kwargs_logs = dict(task_arn            = task_arn                ,
                           task_definition_arn = task_definition_arn     ,
                           image_name          = self.image_name         ,
                           cluster_name        = self.cluster_name )
        return self.ecs().logs(**kwargs_logs)

    def tasks(self):
        return self.ecs().tasks(self.cluster_name)


class Page_Utils:
    pass

    def map_breadcrumbs(self, request):

        breadcrumbs = [
            {'name': 'Home', 'url'  : '#'},
            #{'name': 'Section F'    , 'url': '#'},
            {'name': 'Dev', 'url': None}
        ]
        return breadcrumbs


# from flask import render_template
#
# from cbr_website_beta.aws.cloudwatch.Cloud_Watch_Metrics import Cloud_Watch_Metrics
# from osbot_utils.utils.Misc import list_set, bytes_to_base64
#
# TCB_DASHBOARD_NAME   = 'TCB-Widgets'        # todo: refactor into separate class
#
# class Odin__Panels__Dashboard_Widgets:
#
#     def exposed_methods(self):                                                  # todo: find a better name for this method and move into base class
#         return { 'dashboard_metrics': self.dashboard_metrics,
#                  'widget_screenshot': self.widget_screenshot}
#
#     def dashboard_metrics(self):
#         return render_template(**self.dashboard_metrics__render_kwargs())
#
#     def dashboard_metrics__render_kwargs(self):
#         cloud_watch_metrics = Cloud_Watch_Metrics()
#
#         dashboards_metrics = cloud_watch_metrics.dashboard_widgets(TCB_DASHBOARD_NAME)
#
#         return { "dashboards_metrics"    : dashboards_metrics                      ,
#                  "template_name_or_list" : "dev/cloudwatch/dashboard-metrics.html" ,
#                  "title"                 : "Dashboard - Metrics"                   }
#
#     def widget_screenshot(self, widget_name, start="-PT2H", end="P0D"):
#         cloud_watch_metrics = Cloud_Watch_Metrics()
#         image_png = cloud_watch_metrics.widget_screenshot(TCB_DASHBOARD_NAME, widget_name, start=start, end=end)
#         return image_png
#         #image_base64 = bytes_to_base64(image_png)
#         #return image_base64
#
#
# #
# #
# # def test_dashboard_metrics(self):
# #     dashboard_metrics = self.logs_views.dashboard_metrics()
# #     print(dashboard_metrics)
# #
# #
#



from os import getenv

import openai
from dotenv import load_dotenv
from openai import OpenAI
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.utils.Dev import pprint

OPEN_AI__API_KEY   = 'OPEN_AI__API_KEY'
DEFAULT_MAX_TOKENS = 2048 


class API_Open_AI:

    def __init__(self):
        self.stream              = True
        self.seed                = 42
        self.temperature         = 0.0
        self.model               = 'gpt-4-vision-preview'  # in Nov 2023 is there is a limit of 100 images per day
        #self.model               = 'gpt-4-turbo-preview'     # gpt-3.5-turbo' #'gpt-4' #
        self.print_create_kwargs = True

    @cache_on_self
    def api_key(self):
        load_dotenv()
        return getenv(OPEN_AI__API_KEY)

    def create(self, messages, model=None, temperature=None, seed=None, max_tokens=None):
        openai.api_key = self.api_key()
        kwargs = dict(model       = model       or self.model      ,
                      messages    = messages                       ,
                      temperature = temperature or self.temperature,
                      seed        = seed        or self.seed       ,
                      stream      = self.stream                    ,
                      max_tokens  = max_tokens  or DEFAULT_MAX_TOKENS)

        if self.print_create_kwargs:                            # todo : remove
            pprint(kwargs)
        client = OpenAI(api_key=self.api_key())
        response = client.chat.completions.create(**kwargs)

        return self.parse_response(response)

    def messages(self):
        return [{"role": "user", "content": 'Hi'}]

    def parse_response(self, response):
        for chunk in response:
            choices = chunk.choices
            if len(choices) == 1:       # todo: add support for multiple choices
                choice      = choices[0]
                delta       = choice.delta
                yield delta.content

    def setup(self):
        openai.api_key = self.api_key()
        return self

    def ask_one_question_no_history(self,question, model=None, async_mode=False):
        messages    = [{"role": "user", "content": question}]
        return self.ask_using_messages(messages, model=model, async_mode=async_mode)

    def ask_using_messages(self, messages, model=None, temperature=None, seed=None, max_tokens=None,  async_mode=False):
        generator    = self.create(messages, model=model, temperature=temperature,seed=seed, max_tokens=max_tokens)
        if async_mode:
            return generator
        full_answer = ""

        for item in generator:
            if item is not None:
                full_answer += item
        return full_answer


    def ask_using_system_prompts(self, user_prompt, images=None, system_prompts=None, histories=None, model=None, temperature=None, seed=None, max_tokens=None, async_mode=False):
        messages = []
        if system_prompts:
            for system_prompt in system_prompts:
                messages.append({"role": "system", "content": system_prompt})
        if histories:
            for item in histories:
                question = item.question
                answer   = item.answer
                messages.append({"role": "user"     , "content": question})
                messages.append({"role": "assistant", "content": answer})
        if images:
            user_prompt  = [ { "type"      : "text"      ,
                               "text"      : user_prompt }]
            for image_url in images:
                image = { "type"     : "image_url"       ,
                         "image_url": { "url": image_url }}
                user_prompt.append(image)
        messages.append({"role": "user", "content": user_prompt})

        return self.ask_using_messages(messages, model=model, temperature=temperature, seed=seed, max_tokens=max_tokens,  async_mode=async_mode)

    def ask_question_with_user_data_and_prompt(self,user_question, user_data, system_prompt, user_history):
        messages = [{"role": "system", "content": system_prompt},
                    {"role": "system", "content": user_data    }]
        for item in user_history:
            question = item.get('question')
            answer   = item.get('answer')
            messages.append({"role": "user"     , "content": question})
            messages.append({"role": "assistant", "content": answer})
        messages.append({"role": "user", "content": user_question})

        #pprint(messages)
        return self.ask_using_messages(messages)


from cbr_website_beta.flask.Flask_Site import Flask_Site
from cbr_website_beta.flask.filters.Current_User import client__logged_in
from osbot_playwright.html_parser.Html_Parser import Html_Parser


class Render_Element:
    def __init__(self, path='/', method='GET', data=None, headers=None, element_id=None, tag=None, username=None, user_groups=None, app=None):
        self.method       = method
        self.data         = data
        self.element_id   = element_id
        self.headers      = headers
        self.path         = path
        self.tag          = tag
        self.app          = app or Flask_Site().app()
        if username:
            self.client = client__logged_in(app=self.app, user_name=username, user_groups=user_groups)
        else:
            self.client       = self.app.test_client()
        self.contents     = None
        self.html_parser  = None
        self.raw_element  = None
        self.response     = None


    def __enter__(self):
        self.response    = self.client.open(self.path, method=self.method, data=self.data, headers=self.headers)
        self.html_parser = Html_Parser(self.response.data)
        if self.tag:
            self.raw_element = self.html_parser.soup.find(self.tag)
        else:
            self.raw_element = self.html_parser.soup.find(id=self.element_id)
        if self.raw_element:
            self.contents = Html_Parser(self.raw_element.decode_contents())
        return self, self.contents

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def html(self):
        return self.html_parser.html()

    def title(self):
        return self.html_parser.title()




import traceback

from cbr_website_beta._cbr_shared.schemas.CBR_Logging import CBR_Logging
from cbr_website_beta.apps.CBR_Config import cbr_config
from cbr_website_beta.utils.Web_Utils import Web_Utils
from osbot_aws.aws.dynamo_db.domains.DyDB__Table_With_Timestamp import DyDB__Table_With_Timestamp
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Misc import date_time_now

DYNAMO_DB__TABLE_NAME__CHAT_THREADS = '{env}__cbr_chat_threads'


class DyDB__CBR_Chat_Threads(DyDB__Table_With_Timestamp):

    env: str = cbr_config.env()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.table_name    = DYNAMO_DB__TABLE_NAME__CHAT_THREADS.format(env=self.env)
        self.disabled      = cbr_config.aws_enabled()

    def date_today(self):
        return date_time_now(date_time_format='%Y-%m-%d')

    def documents__today(self):
        if self.disabled:
            return []
        index_name  = 'date'
        index_type  = 'S'
        index_value = date_time_now(date_time_format='%Y-%m-%d')
        documents = self.query_index(index_name=index_name, index_type=index_type, index_value=index_value)
        return documents



dydb_cbr_chat_threads = DyDB__CBR_Chat_Threads()



from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Misc import random_string
from cbr_website_beta.aws.s3.DB_Session import DB_Session


class Temp_DB_Session:
    def __init__(self, data=None, metadata=None, session_id = None, prefix='temp_db_session_'):
        self.session_data     = data
        self.session_id       = session_id or random_string(prefix=prefix)
        self.session_metadata = metadata
        self.db_session       = DB_Session(self.session_id)

    def __enter__(self):
        return self.create()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.delete()
        pass

    def create(self):
        self.db_session.create(data=self.session_data, metadata=self.session_metadata)
        return self.db_session

    def delete(self):
        self.db_session.delete()
        return self

    def exists(self):
        return self.db_session.exists()


from cbr_website_beta.aws.s3.DB_Odin_Data import DB_Odin_Data
from cbr_website_beta.aws.s3.DB_Sessions import DB_Sessions
from cbr_website_beta.aws.s3.DB_Users import DB_Users
from osbot_utils.testing.Duration import Duration
from osbot_utils.utils.Dev import pprint


class S3_Backend_Analysis:

    def __init__(self):
        self.sessions  = DB_Sessions ()
        self.users     = DB_Users    ()
        self.odin_data = DB_Odin_Data()

    def s3_update_db_sessions_status(self):
        sessions_data = self.sessions.sessions__all_data()
        return self.odin_data.current_sessions__save_to_s3(sessions_data)

    def s3_update_db_users_metadata(self):
        users_metadatas = self.users.users__all_metadatas()
        return self.odin_data.users_metadatas__save(users_metadatas)






from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self

# todo : refactor this out of this code since in the future this will need to be a REST call to ODIN
#        who will be responsible for capturing the logs

class CBR_Logging(Kwargs_To_Self):
    # auto populated
    id          : str           # to be set by Dynamo_DB__Table
    timestamp   : int           # to be set by the request
    # indexes
    date        : str   = 'NA'
    level       : str   = 'NA'
    message     : str   = 'NA'
    source      : str   = 'NA'
    topic       : str   = 'NA'

    # other
    city        : str   = 'NA'
    country     : str   = 'NA'
    user_id     : str   = 'NA'
    extra_data  : dict



from flask import render_template, redirect

from cbr_website_beta.apps.CBR_Config                   import cbr_config
from cbr_website_beta.apps.llms.Initial_Message         import Initial_Message
from cbr_website_beta.apps.llms.Prompt_Examples         import Prompt_Examples
from cbr_website_beta.apps.llms.System_Prompt           import System_Prompt
from cbr_website_beta.apps.llms.llms_routes             import user_data_for_prompt, current_user_data, target_athena_url
from cbr_website_beta.apps.root                         import blueprint
from cbr_website_beta.apps.user.user_profile            import render_page__login_required
from cbr_website_beta.flask.decorators.allow_annonymous import allow_anonymous
from cbr_website_beta.flask.utils.current_server        import current_server
from cbr_website_beta.utils.Site_Utils                  import Site_Utils
from cbr_website_beta.utils.Version import Version


@blueprint.route('/')
@allow_anonymous
def index():
    #if Current_User().is_logged_in():
        home_uri = f"{current_server()}home"
        return redirect(home_uri)
    #return render_template('home/accounts/login.html', segment='login')

@blueprint.route('/version')
@allow_anonymous
def version():
    return Version().value()

@blueprint.route('/home')
@blueprint.route('/home.html')
@allow_anonymous
def home():
    user_data       = current_user_data()
    title           = 'Welcome'
    first_name      = user_data.get('First name','')
    last_name       = user_data.get('Last name' ,'')
    content_view    = 'includes/home.html'

    template_name = '/pages/page_with_view.html'
    return render_template(template_name_or_list = template_name,
                           title                 =  title       ,
                           content_view          = content_view ,
                           first_name            = first_name   ,
                           last_name             = last_name    )

@blueprint.route('/athena')
@allow_anonymous
def athena():
    user_data = user_data_for_prompt()
    title     = 'Athena'
    if user_data or cbr_config.login_disabled():
        url_athena       = target_athena_url() + '/open_ai/prompt_with_system__stream'  # todo: refactor into helper method
        content_view     = '/llms/open_ai/chat_bot_ui.html'
        template_name    = '/pages/page_with_view.html'
        examples_title   = 'Prompt examples'

        return render_template( template_name_or_list = template_name               ,
                                content_view          = content_view                ,
                                examples_texts        = Prompt_Examples().athena()  ,        # todo: refactor to not need to call Prompt_Examples() on all calls
                                examples_title        = examples_title              ,
                                initial_message       = Initial_Message().athena()  ,         # todo: refactor to not need to call Prompt_Examples() on all calls
                                system_prompt         = System_Prompt().athena()    ,
                                title                 = title                       ,
                                url_athena            = url_athena                  ,
                                user_data             = user_data                   )
    else:
        return render_page__login_required(title)


# import functools
# import sys
# #from aws_xray_sdk.core import patch_all, xray_recorder
# from cbr_website_beta.flask.middleware.xray_middleware import current_segment_save_to, current_segment_print
#
#
# def xray_trace_segment(name="Xray Test", print_traces=False, save_json_to=None):
#
#     def decorator(f):
#         @functools.wraps(f)
#         def wrapper(*args, **kwargs):
#             subsegment = xray_recorder.begin_segment(name)
#             patch_all()
#             try:
#                 result = f(*args, **kwargs)
#             except Exception as e:
#                 exc_type, exc_value, exc_traceback = sys.exc_info()
#                 subsegment.add_exception(exc_value, exc_traceback)
#                 raise
#             finally:
#                 if save_json_to:
#                     current_segment_save_to(save_json_to)
#                 if print_traces:
#                     current_segment_print()
#                 xray_recorder.end_segment()
#             return result
#         return wrapper
#     return decorator


import osbot_llms
from osbot_utils.utils.Files import path_combine, file_contents


class API_OSBot_LLMs:

    def path_osbot_llms(self):
        return osbot_llms.path

    def path_version_file(self):
        return path_combine(self.path_osbot_llms(), 'version')

    def path_web_static(self):
        return path_combine(self.path_osbot_llms(), 'web_static/src')

    def url_path_chatbot_js(self):
        return '/llms-ui/chatbot_openai/Chatbot_OpenAI.js'

    def version(self):
        return file_contents(self.path_version_file()).strip()


from flask import render_template

from cbr_website_beta._cbr_shared.dynamo_db.DyDB__CBR_Chat_Threads import DyDB__CBR_Chat_Threads
from cbr_website_beta._cbr_shared.dynamo_db.DyDB__CBR_Logging import DyDB__CBR_Logging
from cbr_website_beta._cbr_shared.dynamo_db.DyDB__CBR_Requests import DyDB__CBR_Requests
from cbr_website_beta._cbr_shared.dynamo_db.DyDB__CBR_User_Sessions import DyDB__CBR_User_Sessions
from cbr_website_beta.apps.CBR_Config import cbr_config
from cbr_website_beta.apps.dev.views.Logs_Views import Logs_Views
from cbr_website_beta.utils.Web_Utils import Web_Utils
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Lists import list_remove
from osbot_utils.utils.Misc import date_time_now

CHAT_THREAD__GPT_RESPONSE__MAX_SIZE = 250

class Odin__Panels__Logs:

    def __init__(self):
        self.logs_view = Logs_Views()

    def exposed_methods(self):                                                  # todo: find a better name for this method and move into base class
        return { 'cbr_chat_threads' : self.cbr_chat_threads ,
                 'cbr_logging'      : self.cbr_logging      ,
                 'cbr_requests'     : self.cbr_requests     ,
                 'cbr_user_sessions': self.cbr_user_sessions,
                 'ip_address'       : self.ip_address       }

    # exposed methods
    def cbr_chat_threads (self, index_name=None, index_value=None, env:str="PROD", hours:int=1):
        kwargs        = dict(index_name=index_name, index_value=index_value, env=env, hours=hours)
        render_kwargs = self.cbr_chat_threads__render_kwargs(**kwargs)
        return render_template(**render_kwargs)

    def cbr_logging (self, index_name=None, index_value=None, env:str="PROD", hours:int=1):
        kwargs        = dict(index_name=index_name, index_value=index_value, env=env, hours=hours)
        render_kwargs = self.cbr_logging__render_kwargs(**kwargs)
        return render_template(**render_kwargs)

    def cbr_requests(self, index_name=None, index_value=None, env:str="PROD", hours:int=1):
        kwargs = dict(index_name=index_name, index_value=index_value, env=env, hours=hours)
        return render_template(**self.cbr_requests__render_kwargs(**kwargs))

    def cbr_user_sessions (self, index_name=None, index_value=None, env:str="PROD", hours:int=1):
        kwargs        = dict(index_name=index_name, index_value=index_value, env=env, hours=hours)
        render_kwargs = self.cbr_user_sessions__render_kwargs(**kwargs)
        return render_template(**render_kwargs)

    def ip_address(self, ip_address: str='NA', env:str="PROD", hours:int=1):
        kwargs = dict(ip_address=ip_address, env=env, hours=hours)
        return render_template(**self.ip_address__render_kwargs(**kwargs))

    # internal methods

    def cbr_chat_threads__render_kwargs(self, index_name=None, index_value=None, env=None, hours=5):
        env                   = env or cbr_config.env()
        dydb_cbr_chat_threads = DyDB__CBR_Chat_Threads(env=env)
        fields                = ['id', 'session_id', 'when', 'user_prompt', 'gpt_response', 'source', 'chat_thread_id', 'date']
        if not index_name:
            index_name = 'date'
            index_value = date_time_now(date_time_format='%Y-%m-%d')

        requests = self.logs_view.api_log_data(dydb=dydb_cbr_chat_threads, index_name=index_name, index_value=index_value,
                                               hours=hours, env=env, fields=fields)

        for request in requests:
            gpt_response = request.get('gpt_response') or ''
            if len(gpt_response) > CHAT_THREAD__GPT_RESPONSE__MAX_SIZE:
                request['gpt_response'] = gpt_response[:CHAT_THREAD__GPT_RESPONSE__MAX_SIZE] + f' ... (size {len(gpt_response)})'
            session_id = request.get('session_id') or ''
            if len(session_id.split('__')) ==2 :
                user_name = session_id.split('__')[0]
                request['session_id'] = user_name
        metadata = dict(fields=fields, hours=hours, env=env)
        render_kwargs = dict(title                 = 'Chat Threads'                     ,
                             requests              = requests                           ,
                             metadata              = metadata                           ,
                             method_name           = 'cbr_chat_threads'                 ,
                             template_name_or_list = "dev/odin/panels/cbr_requests.html")
        return render_kwargs


    def cbr_logging__render_kwargs(self, index_name=None, index_value=None, env=None, hours=5):
        env               = env or cbr_config.env()
        dydb_cbr_requests = DyDB__CBR_Logging(env=env)
        load_fields       = [ 'id'         ,
                              'source'     , 'when'        , 'message'    , 'level'      ,
                              'topic'      , 'user_id'     , 'extra_data' , 'env'        ]
        view_fields       = [ 'id'         ,
                              'source'     , 'when'        , 'message'    ,
                              'level'      , 'topic'       , 'env'        ]

        if not index_name:
            index_name = 'date'
            index_value = date_time_now(date_time_format='%Y-%m-%d')
        requests = self.logs_view.api_log_data(dydb=dydb_cbr_requests, index_name=index_name, index_value=index_value, hours=hours, env=env,fields=load_fields)

        self.filter_logging_requests(requests=requests)
        metadata = dict(fields=view_fields, hours=hours, env=env)
        render_kwargs = dict(title                 = 'Odin - Logs'  ,
                             requests              = requests       ,
                             metadata              = metadata       ,
                             method_name           = 'cbr_logging'  ,
                             template_name_or_list = "dev/odin/panels/cbr_requests.html")

        return render_kwargs

    def filter_logging_requests(self, requests):
        for request in requests:
            extra_data = request.get('extra_data', {})
            message    = request.get('message'   , '')
            if message == 'prompt_request':
                user_prompt  = extra_data.get('user_prompt' , '')
                histories    = extra_data.get('prompt_data' , {}).get('histories') or []
                gtp_response = extra_data.get('gtp_response', '')
                if len(user_prompt)  > 300:
                    user_prompt = user_prompt[:300] + f'<i> ... (hidden={len(user_prompt)-300})</i>'
                if len(gtp_response) > 500:
                    gtp_response = gtp_response[:500] + f'<i> ... (hidden={len(gtp_response)-500})</i>'
                request['user_prompt' ] =user_prompt
                request['gtp_response'] = gtp_response
                request['history'     ] = len(histories)
            else:
                request['user_prompt' ] = ''
                request['gtp_response'] = ''
                request['history'     ] = 0


    def cbr_requests__render_kwargs(self, index_name=None, index_value=None, env=None, hours=1):
        env               = env or cbr_config.env()
        dydb_cbr_requests = DyDB__CBR_Requests(env=env)
        fields   = ['id',
                    'source'     , 'when'       , 'user_name'  , 'path'       , 'duration'   ,'city', 'country' ,
                    'ip_address' ,'ip_info'     , 'session_id' , 'level'      ,'referer',
                    'status_code', 'env'        ]
                    # 'date',  'user_role' ,'method'     ,
                    # 'host',
                    #'req_id',
                    #, 'user_status'

        ignore = dict(path='/dev/dev-panel')

        if not index_name:
            index_name = 'date'
            index_value = date_time_now(date_time_format='%Y-%m-%d')
        requests = self.logs_view.api_log_data(dydb=dydb_cbr_requests, index_name=index_name, index_value=index_value, hours=hours, env=env, fields=fields, ignore=ignore)

        metadata = dict(fields=fields, hours=hours, env=env)
        render_kwargs = dict(title                 = 'Web/APIs Requests',
                             requests              = requests           ,
                             metadata              = metadata           ,
                             method_name           = 'cbr_requests'     ,
                             template_name_or_list = "dev/odin/panels/cbr_requests.html" )

        return render_kwargs

    def cbr_user_sessions__render_kwargs(self, index_name=None, index_value=None, env=None, hours=5):
        env                    = env or cbr_config.env()
        dydb_cbr_user_sessions = DyDB__CBR_User_Sessions(env=env)
        fields                 = ['id', 'user_name', 'display_name', 'session_id', 'source', 'date']
        if not index_name:
            index_name = 'date'
            index_value = date_time_now(date_time_format='%Y-%m-%d')

        requests = self.logs_view.api_log_data(dydb=dydb_cbr_user_sessions, index_name=index_name, index_value=index_value,
                                               hours=hours, env=env, fields=fields)

        metadata = dict(fields=fields, hours=hours, env=env)
        render_kwargs = dict(title                 = 'User Sessions'                    ,
                             requests              = requests                           ,
                             metadata              = metadata                           ,
                             method_name           = 'cbr_user_sessions'                ,
                             template_name_or_list = "dev/odin/panels/cbr_requests.html")
        return render_kwargs

    def ip_address__render_kwargs(self, ip_address, env, hours):
        ip_address_data = self.logs_view.get_ip_address_data(ip_address)
        render_kwargs = dict(env                   = env              ,
                             ip_address            = ip_address       ,
                             ip_address_data       = ip_address_data  ,
                             hours                 = hours            ,
                             template_name_or_list ="dev/odin/panels/ip_address.html")
        return render_kwargs



from os import environ

from cbr_website_beta.llms.API_Open_AI import API_Open_AI
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.helpers.sqlite.Sqlite__Database import Sqlite__Database
from osbot_utils.helpers.sqlite.domains.Sqlite__DB__Json import Sqlite__DB__Json
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Files import current_temp_folder, path_combine
from osbot_utils.utils.Json import json_dump, json_dumps, json_loads
from osbot_utils.utils.Misc import str_sha256, timestamp_utc_now

ENV_NAME_PATH_LOCAL_DBS      = 'PATH_LOCAL_DBS'
DB_NAME                      = 'open_ai_cache.sqlite'
TABLE_NAME__OPEN_AI_REQUESTS = 'open_ai_requests'

class Table__Open_AI__Requests(Kwargs_To_Self):
    request_data  : str
    request_hash  : str
    response_data : str
    response_hash : str
    timestamp     : int
    comment       : str

class Open_AI__Cache:
    #sqlite_db_json : Sqlite__DB__Json
    sqlite_database: Sqlite__Database

    def __init__(self) -> None:
        self.sqlite_database = Sqlite__Database(db_path=self.path_db())

    @cache_on_self
    def api_open_ai(self):
        return API_Open_AI()

    def cached_data(self, **kwargs):
        table = self.table_open_ai_requests()
        request_data = json_dump(kwargs)
        request_hash = str_sha256(request_data)
        cached_data = table.select_row_where(request_hash=request_hash)
        return cached_data

    def cached_response(self, **kwargs):
        cached_data = self.cached_data(**kwargs)
        if cached_data:
            response_data_str = cached_data.get('response_data')
            response_data = json_loads(response_data_str)
            return response_data
        return None

    def call_api_open_ai__create(self,**kwargs):   # messages, model=None, temperature=None, seed=None, max_tokens=None
        return self.api_open_ai().create(**kwargs)

    def call_chat_completion_create(self, **kwargs):
        table = self.table_open_ai_requests()
        cached_data = self.cached_data(**kwargs)
        if cached_data:
            response_data_str = cached_data.get('response_data')
            response_data     = json_loads(response_data_str)
        else:
            request_data = json_dump(kwargs)                            # todo refactor this code so that we don't have to calculate this twice
            request_hash = str_sha256(request_data)
            chunks = []
            response_data_stream = self.call_api_open_ai__create(**kwargs)
            for chunk in response_data_stream:
                if chunk:
                    chunks.append(chunk)
            response_data = json_dumps(chunks)
            response_hash = str_sha256(response_data)
            row_data = { 'comment'      : ''                 ,
                         'request_data'  : request_data       ,
                         'request_hash'  : request_hash       ,
                         'response_data' : response_data      ,
                         'response_hash' : response_hash      ,
                         'timestamp'     : timestamp_utc_now()}
            table.add_row_and_commit(**row_data)                                # add to cache


        return response_data

    def path_db_folder(self):
        return environ.get(ENV_NAME_PATH_LOCAL_DBS) or current_temp_folder()

    def path_db(self):
        return path_combine(self.path_db_folder(), DB_NAME)

    def table_open_ai_requests(self):
        table = self.sqlite_database.table(TABLE_NAME__OPEN_AI_REQUESTS)
        table.row_schema = Table__Open_AI__Requests
        if table.not_exists():
            table.create()
        return table



import sys

import cbr_website_beta

from osbot_utils.utils.Files import path_combine, file_contents
from osbot_utils.utils.Json import json_file_load, json_file_create


class Site_Utils:

    def path_code_root(self):
        return cbr_website_beta.__path__[0]

    def path_source_code_root(self):
        return path_combine(self.path_code_root(), '..')

    @staticmethod
    def running_in_pytest():
        return 'pytest' in sys.modules



import cbr_website_beta
from cbr_website_beta.utils.Version import version
from osbot_utils.decorators.methods.cache_on_self   import cache_on_self
from osbot_utils.utils.Status                       import status_ok, status_error
from osbot_utils.utils.Files                        import path_combine
from osbot_utils.utils.Env import get_env, load_dotenv
from osbot_utils.base_classes.Type_Safe             import Type_Safe
from osbot_utils.utils.Toml                         import toml_dict_from_file

ENV_VAR_NAME__CBR_CONFIG_FILE     = 'CBR_CONFIG_FILE'
DEFAULT__CBR_CONFIG_FILE          = 'cbr-website.offline.toml'      # default to the offline mode
FOLDER__CBR_CONFIG_FILES          = '../config'

class CBR_Config(Type_Safe):

    def __init__(self):
        super().__init__()
        load_dotenv()

    @cache_on_self
    def config(self):
        try:
            config_data = toml_dict_from_file(self.path_config_file())
            return status_ok(data=config_data)
        except Exception as error:
            return status_error(message='In CBR_Config failed to load config file', error=f'{error}')

    def current_config_file_name(self):
        return get_env(ENV_VAR_NAME__CBR_CONFIG_FILE, DEFAULT__CBR_CONFIG_FILE)

    def path_config_file(self):
        return path_combine(self.path_config_files(), self.current_config_file_name())

    def path_config_files(self):
        return path_combine(cbr_website_beta.path, FOLDER__CBR_CONFIG_FILES)

    # top level sections

    def cbr_dev(self):
        return self.config().get('data',{}).get('cbr_dev', {})

    def cbr_website(self):
        return self.config().get('data',{}).get('cbr_website', {})

    # from env variables

    def env(self):
        return get_env('EXECUTION_ENV', 'LOCAL')


    # specific values

    def assets_dist(self):
        if self.env() == 'LOCAL':
            value = '/dist'
        else:
            value = f'https://static.thecyberboardroom.com/dist/{version}'
        return value

    def assets_root(self):
        return self.cbr_website().get('assets_root')

    def aws_enabled(self):
        return self.cbr_website().get('aws_enabled')

    def aws_disabled(self):
        return self.aws_enabled() is False

    def dev__capture_exceptions(self):
        return self.cbr_dev().get('capture_exceptions', True)

    def login_enabled(self):
        return self.cbr_website().get('login_enabled')

    def login_disabled(self):
        return self.login_enabled() is False

    # static values  # todo: figure out best place to store these, since they really shouldn't change
    def session_cookie_httponly(self):
        return True

    def remember_cookie_httponly(self):
        return True

    def remember_cookie_duration(self):
        return 3600                         # todo: look at increasing this value, since I think this is reason why the user session expires quite often

cbr_config = CBR_Config()



import requests

from cbr_website_beta.bots.Athena_Rest_API import Athena_Rest_API
from cbr_website_beta.bots.schemas.Create_User_Session import Create_User_Session
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Json import json_dumps

from osbot_utils.utils.Misc import str_to_base64

from cbr_website_beta.flask.utils.current_server import current_server


class Cognito_Auth_Flow:

    def convert_auth_code_to_auth_data(self, sign_in_code, redirect_uri):
        project     = 'the-cbr-beta.auth'
        region      = "eu-west-2"
        client_id   = "5ij6l5kdho4umoks5rjfh9cbid"
        grant_type  = "authorization_code"

        token_url = f"https://{project}.{region}.amazoncognito.com/oauth2/token"

        payload = {
            "grant_type"  : grant_type,
            "client_id"   : client_id            ,
            "code"        : sign_in_code,
            "redirect_uri": redirect_uri
        }
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        response = requests.post(token_url, data=payload, headers=headers)

        return response.json()

    def create_cbr_token_cookie_from_cognito_code(self, sign_in_code):
        #print('-------- create_cbr_token_cookie_from_cognito_code ----------  ')
        try:
            redirect_uri   = f"{current_server()}sign-in"
            cognito_tokens = self.convert_auth_code_to_auth_data(sign_in_code, redirect_uri)
            access_token   = cognito_tokens.get('access_token')
            user_info      = self.parse_jwt_token(access_token)
            metadata       = {'source': 'cognito'}
            db_session    = self.create_session(user_info, metadata)
            session_id     = self.create_session_in_athena(user_info, cognito_tokens)          # create the session and use the current document id as the cookie session id
            cookie_value   = session_id
            #cookie_value  = db_session.session_id                                               # give the session_id to the user as a cookie
            return {'cookie_value': cookie_value, 'user_info': user_info}
        except Exception as error:  # todo : add proper error handling, namely redirecting to an error 500 page
            # todo: add logging of error
            return None

    def create_session(self, user_info, metadata=None):
        from cbr_website_beta.aws.s3.DB_Sessions import DB_Sessions
        db_sessions = DB_Sessions()
        user_name   = user_info.get('username')
        user_jti    = user_info.get('jti')
        session_id  = f"{user_name}__{user_jti}"
        db_session  = db_sessions.db_session(session_id)
        db_session.create(user_info, metadata)
        return db_session

    def create_session_in_athena(self, user_info, cognito_tokens):
        user_name       = user_info.get('username')
        session_id      = user_info.get('jti')
        source          = 'CognitoAuth'
        athena_rest_api = Athena_Rest_API()
        kwargs = dict(user_name     = user_name ,
                      session_id    = session_id,
                      source        = source   ,
                      cognito_tokens = cognito_tokens)
        create_user_session = Create_User_Session(**kwargs)
        response             = athena_rest_api.user__create_session(create_user_session)
        session_id           = response.get('session_id')
        return session_id



    def parse_jwt_token(self, access_token):
        import jwt
        decoded_access_token = jwt.decode(access_token, algorithms=["RS256"], options={"verify_signature": False})
        return decoded_access_token



from cbr_website_beta._testing.test_utils.TestCase__CBR__Website import TestCase__CBR__Website


class CBR__API_Pages(TestCase__CBR__Website):

    # API pages
    def api_user_data(self): return self.open_json('/api/user-data')


# todo: find a better solution to get the DNS name of the current server, since at the moment
#       it requests an hard-coded value added to CloudFront (Cloudfront-Domain)
from flask import request


def current_server():
    cloud_front_domain = request.headers.get('Cloudfront-Domain')
    if cloud_front_domain:
        current_server = cloud_front_domain
    else:
        current_server = request.url_root
    if 'http://' in current_server:                                                             # todo: figure out better solution for this (currently seen when proxying traffic via ngrok)
        if current_server.startswith('http://localhost') is False:                              # can't do this for localhost
            print(f'note: need to fix current_server since it is http: {current_server}')       # todo: add to main logging class (when wired up to DynamoDB)
            current_server = current_server.replace('http://', 'https://')
    return current_server


# from functools import wraps
# from flask import redirect, url_for
#
# from flask import request
#
# from cbr_website_beta.apps.root import blueprint
# from cbr_website_beta.flask.filters.Current_User import Current_User, USER_DATA_WITH_NO_CBR_TOKEN
#
# def get_current_user_from_cookie():
#     return Current_User().current_user(context=None, field='username')
#
#
# def login_required(f):
#     @wraps(f)
#     def decorated_function(*args, **kwargs):
#         user = get_current_user_from_cookie()
#         if user is USER_DATA_WITH_NO_CBR_TOKEN:
#             return redirect(url_for('user_blueprint.unauthorized'))
#         return f(*args, **kwargs)
#     return decorated_function



# -*- encoding: utf-8 -*-
import os

from werkzeug.exceptions import NotFound

from cbr_website_beta._cbr_shared.dynamo_db.DyDB__CBR_Logging import DyDB__CBR_Logging
from cbr_website_beta.apps.llms.llms_routes import target_athena_url
from cbr_website_beta.flask.decorators.allow_annonymous import allow_anonymous
from cbr_website_beta.llms.API_OSBot_LLMs import API_OSBot_LLMs
from cbr_website_beta.utils.Page_Utils import Page_Utils
from cbr_website_beta.apps.home import blueprint
from flask import render_template, request, send_from_directory, make_response
from jinja2 import TemplateNotFound

from cbr_website_beta.utils.Site_Utils import Site_Utils
from cbr_website_beta.utils.Version import version

COGNITO_SIGN_IN       = 'https://the-cbr-beta.auth.eu-west-2.amazoncognito.com/login?client_id=5ij6l5kdho4umoks5rjfh9cbid&response_type=code&scope=email+openid+phone&redirect_uri={cbr_domain}/sign-in'
COGNITO_SIGN_OUT      = 'https://the-cbr-beta.auth.eu-west-2.amazoncognito.com/logout?client_id=5ij6l5kdho4umoks5rjfh9cbid&response_type=code&scope=email+openid+phone&logout_uri={cbr_domain}/sign-out'
CBR_DOMAIN            = 'https://www.thecyberboardroom.com'
#DEV_DOMAIN            = 'http://localhost:5000'
EXPECTED_ROUTES__HOME = [ '/<path:path>' , '/llms-ui/<path:path>', '/terms_and_conditions', '/service-worker.js']

EXPECTED_METHODS_HOME = ['home_blueprint.llms_web_static'           ,
                         'home_blueprint.path_path'                 ,
                         'home_blueprint.service_worker'            ,
                         'home_blueprint.terms_and_conditions'      ,
                         'llms_blueprint.just_chat_ui'              ,
                         'llms_blueprint.multiple_athenas'          ,
                         'llms_blueprint.no_system_prompt'          ]

dydb_cbr_logging = DyDB__CBR_Logging()
def running_in_aws():
    return 'AWS_EXECUTION_ENV' in os.environ

# @blueprint.route('/static/assets/<path:path>')
# @allow_anonymous
# def static_assets(path):
#     if running_in_aws():
#         folder_assets = '/tmp/static/assets'
#     else:
#         folder_assets = '../../flask_assets/static/assets'
#     return send_from_directory(folder_assets, path)

path_osbot_llms_web_static = API_OSBot_LLMs().path_web_static()

@blueprint.route('/service-worker.js')
@allow_anonymous
def service_worker():
    content  = render_template('scripts/service-worker.js', version=version)
    response = make_response(content)
    response.headers['Content-Type'] = 'application/javascript'
    return response


@blueprint.route('/llms-ui/<path:path>')
@allow_anonymous
def llms_web_static(path):
    try:
        return send_from_directory(path_osbot_llms_web_static, path)
    except NotFound as notfound:
        return "File not found", 404


@blueprint.route('/terms_and_conditions')
@allow_anonymous
def terms_and_conditions():
    return render_template('home/terms_and_conditions.html')

@blueprint.route('/<path:path>')
@allow_anonymous
def path_path(path):            # todo: remove the need for this route
    try:
        #if path == 'athena'  : path = 'athena/index.html'           # todo: find a better to map these top level paths
        if path == 'content' : path = 'content/index.html'
        if path.endswith("/"): path += "index.html"
        elif not path.endswith('.html'):    path += '.html'

        # Detect the current page
        kwargs = dict(template_name_or_list = "home/" + path                                             ,
                      breadcrumbs           = map_breadcrumbs(request)                                   ,
                      url_athena            = target_athena_url() + '/open_ai/prompt_with_system__stream',  # todo: refactor direct render of the multiple athena UIs
                      page_title            = 'title'                                                    )

        return render_template(**kwargs)

    except TemplateNotFound as error:
        print(f"***** error 404 : {error}")
        return render_template('home/page-404.html', path=path), 404

    except Exception as error:
        print(f"***** error 500 : {error}")
        dydb_cbr_logging.log_exception(error)
        return render_template('home/page-500.html'), 500


# Helper - Extract current page name from request
def map_breadcrumbs(request):

    return Page_Utils().map_breadcrumbs(request)
    # breadcrumbs = [
    #     # {'name': 'Home', 'url'  : '#'},
    #     # {'name': 'Section A'    , 'url': '#'},
    #     # {'name': 'Blank Page!!!', 'url': None}
    # ]
    # path = ''
    # for key in request.path.split('/'):
    #     path += '/{key}'
    #     breadcrumbs.append({'name': key, 'url': path})
    # pprint(breadcrumbs)
    # return breadcrumbs
    # try:
    #     segment = request.path.split('/')[-1]
    #     if segment == '':
    #         segment = 'index'
    #     return segment
    # except:
    #     return None



from cbr_website_beta._testing.test_utils.TestCase__CBR__Website import TestCase__CBR__Website
from cbr_website_beta.apps.home.home_routes import COGNITO_SIGN_IN


class CBR__Pages(TestCase__CBR__Website):

    # CBR Pages
    def athena   (self): return self.open('/athena'         )
    def content  (self): return self.open('/content'        )
    def login    (self): return self.open('/login'          )
    def home     (self): return self.open('/home'           )
    def videos   (self): return self.open('/content/videos' )







from osbot_utils.utils.Files import folder_name, parent_folder


def convert_paths_into_folder_dict(data_set):
    def insert_into_dict(d, parts):
        if len(parts) == 1:
            if "files" not in d:
                d["files"] = {}
            d["files"][parts[0]] = None  # Or some default value
        else:
            head, *tail = parts
            if head not in d:
                d[head] = {}
            insert_into_dict(d[head], tail)

    folder_dict = {}

    for item in data_set:
        parts = item.split('/')
        insert_into_dict(folder_dict, parts)

    return folder_dict

def parent_folder_name(target):
    return folder_name(parent_folder(target))


from urllib.parse import urljoin

import requests

from cbr_website_beta.apps.llms.llms_routes import target_athena_url
from cbr_website_beta.bots.schemas.Log_Entry import Log_Entry

PATH__ROOT                        = '/'
PATH__DOCS                        = '/docs'
PATH__OPEN_API_JSON               = '/openapi.json'
PATH__CONFIG__VERSION             = '/config/version'
PATH__CONFIG__CYBER_SECURE_ORG    = '/content/building_a_cybersecure_organisation'
PATH__CONFIG__CYBER_IN_BOARD      = '/content/cybersecurity_in_the_boardroom'
PATH__CONFIG__DIGITAL_TRUST       = "/content/importance_of_digital_trust"
PATH__CONFIG__INCIDENT_MANAGEMENT = '/content/incident_management'
PATH__LOGGING__ADD_LOG_ENTRY      = '/logging/add_log_entry'
PATH__OPENAI__PROMPT_WITH_STREAM  = '/open_ai/prompt_with_system__stream'
PATH__USER__CREATE_USER_SESSION   = '/user/create_user_session'

class Athena_Rest_API:

    def __init__(self):
        self.url = target_athena_url()

    # utils
    def requests_get(self, path):
        url = urljoin(self.url, path)
        response = requests.get(url)
        if response.headers.get('Content-Type') == 'application/json':
            return response.json()
        return response.text

    def requests_post(self, path, data):
        url = urljoin(self.url, path)
        response = requests.post(url, json=data)
        if response.headers.get('Content-Type') == 'application/json':
            return response.json()
        return response.text
    # methods
    #def athena_prompt   (self): return self.requests_get('/athena_prompt'   )
    #def first_question  (self): return self.requests_get('/first_question'  )
    #def git_repo_status (self): return self.requests_get('/git_repo_status' )
    def open_api        (self): return self.requests_get('/openapi.json'    )
    #def ping            (self): return self.requests_get('/ping'            )
    def root            (self): return self.requests_get('/'                )
    def version         (self): return self.requests_get('/config/version'  )

    def content__building_a_cyber_secure_organisation(self):
        return self.requests_get(PATH__CONFIG__CYBER_SECURE_ORG)

    def open_ai__prompt_with_stream(self, data ):
        return self.requests_post(PATH__OPENAI__PROMPT_WITH_STREAM, data)

    def logging__send_log_entry(self, log_entry: Log_Entry):
        json_data = log_entry.dict()
        return self.requests_post(PATH__LOGGING__ADD_LOG_ENTRY, json_data)

    def user__create_session(self, create_user_session):
        json_data = create_user_session.dict()
        return self.requests_post(PATH__USER__CREATE_USER_SESSION, json_data)


from functools import cache

import requests
from cbr_website_beta.aws.s3.S3                     import S3
from osbot_aws.AWS_Config                           import AWS_Config
from osbot_utils.decorators.methods.cache_on_self   import cache_on_self
from osbot_utils.utils.Dev                          import pprint
from osbot_utils.utils.Json                         import json_dumps, json_parse
from osbot_utils.utils.Misc import random_guid, timestamp_utc_now

BUCKET_NAME__DB_USERS        = "{account_id}-db-users"
S3_FOLDER__USERS_METADATA    = 'users_metadata'
S3_FOLDER__USERS_SESSIONS    = 'users_sessions'
S3_FOLDER__ODIN_DATA         = 'odin_data'
S3_FOLDER__TEMP_FILE_UPLOADS = 'temp_file_uploads'

class S3_DB_Base:

    def __init__(self):
        self.aws_config = AWS_Config()

    @cache_on_self
    def s3(self):
        return S3()

    @cache
    def s3_bucket(self):
        return BUCKET_NAME__DB_USERS.format(account_id=self.aws_config.account_id())

    def s3_bucket__temp_data(self):
        return self.aws_config.temp_data_bucket()

    def s3_file_contents(self, s3_key):
        try:
            return self.s3().file_contents(self.s3_bucket(), s3_key)
        except Exception:
            return {}

    def s3_file_data(self, s3_key):
        return json_parse(self.s3_file_contents(s3_key))

    def s3_file_exists(self, s3_key):
        bucket = self.s3_bucket()
        return self.s3().file_exists(bucket, s3_key)

    def s3_file_delete(self, s3_key):
        kwargs = dict(bucket = self.s3_bucket(),
                      key    = s3_key          )
        return self.s3().file_delete(**kwargs)

    def s3_folder_contents(self, folder, return_full_path=False):
        return self.s3().folder_contents(s3_bucket=self.s3_bucket(), parent_folder=folder, return_full_path=return_full_path)

    def s3_folder_files(self, folder, return_full_path=False):
        return self.s3().folder_files(s3_bucket=self.s3_bucket(), parent_folder=folder, return_full_path=return_full_path)

    def s3_folder_list(self, folder, return_full_path=False):
        return self.s3().folder_list(s3_bucket=self.s3_bucket(), parent_folder=folder, return_full_path=return_full_path)

    def s3_save_data(self, data, s3_key):
        data_as_str = json_dumps(data)
        kwargs = dict(file_contents = data_as_str     ,
                      bucket        = self.s3_bucket(),
                      key           = s3_key          )
        return self.s3().file_create_from_string(**kwargs)

    def s3_temp_folder__pre_signed_urls_for_object(self, source='NA', reason='NA', who='NA', expiration=3600):
        s3_bucket          = self.s3_bucket__temp_data()
        s3_temp_folder     = self.s3_folder_temp_file_uploads()
        s3_object_name     = random_guid()
        s3_key             = f'{s3_temp_folder}/{s3_object_name}'
        pre_signed_url__get = self.s3_temp_folder__pre_signed_url(s3_bucket, s3_key, operation='get_object', expiration=expiration)
        pre_signed_url__put = self.s3_temp_folder__pre_signed_url(s3_bucket, s3_key, operation='put_object', expiration=expiration)
        pre_signed_data = dict(pre_signed_url__get = pre_signed_url__get ,
                               pre_signed_url__put = pre_signed_url__put ,
                               reason              = reason              ,
                               timestamp           = timestamp_utc_now() ,
                               source              = source              ,
                               who                 = who                 )
        return pre_signed_data

    def s3_temp_folder__pre_signed_url(self, s3_bucket, s3_key, operation,expiration=3600):
        create_kwargs = dict(bucket_name=s3_bucket,
                             object_name=s3_key,
                             operation=operation,
                             expiration=expiration)
        pre_signed_url = self.s3().create_pre_signed_url(**create_kwargs)
        return pre_signed_url

    def s3_temp_folder__download_string(self, pre_signed_url):
        response = requests.get(pre_signed_url)
        if response.status_code == 200:
            return response.text
        pprint(response)

    def s3_temp_folder__upload_string(self, pre_signed_url, file_contents):
        response = requests.put(pre_signed_url, data=file_contents)
        if response.status_code == 200:
            return True
        else:
            return False

    def setup(self):
        bucket_name = self.s3_bucket()
        if self.s3().bucket_not_exists(bucket_name):
            kwargs = dict(bucket = bucket_name                ,
                          region = self.aws_config.region_name())
            assert self.s3().bucket_create(**kwargs).get('status') == 'ok'
        return True


    def s3_folder_users_sessions(self):
        return S3_FOLDER__USERS_SESSIONS

    def s3_folder_odin_data(self):
        return S3_FOLDER__ODIN_DATA

    def s3_folder_users_metadata(self):
        return S3_FOLDER__USERS_METADATA

    def s3_folder_temp_file_uploads(self):
        return S3_FOLDER__TEMP_FILE_UPLOADS



from flask import render_template
from markdown import Markdown

from cbr_website_beta.data.odin.Analysis__Chat_Threads import Analysis__Chat_Threads

HTML_TITLE__ALL_CHAT_THREADS = 'All Chat Threads'

class Odin__Panels__Chat_Threads:

    def exposed_methods(self):
        return { 'all_chat_threads': self.all_chat_threads ,
                 'chat_thread'     : self.chat_thread      }

    def all_chat_threads(self):
        return render_template(**self.all_chat_threads__kwargs())

    def all_chat_threads__kwargs(self):
        chat_threads = Analysis__Chat_Threads(max_answer_size=200)
        users_threads = chat_threads.users__consolidated_threads()
        return  { "template_name_or_list" : "dev/chat_threads/all_threads.html" ,
                  "title"                 : HTML_TITLE__ALL_CHAT_THREADS        ,
                  'users_threads'         : users_threads                       }

    def chat_thread(self, thread_id):
        chat_threads = Analysis__Chat_Threads(max_answer_size=0)
        thread_data = chat_threads.create_consolidated_thread(thread_id)
        kwargs      = dict(thread_id=thread_id, thread_data=thread_data)
        return render_template(**self.chat_thread__kwargs(**kwargs))

    def chat_thread__kwargs(self, **kwargs):
        return  { "template_name_or_list" : "dev/chat_threads/dev-thread-view.html" ,
                  "title"                 : HTML_TITLE__ALL_CHAT_THREADS            ,
                  **kwargs}




# import requests
# import xml.etree.ElementTree as ET
#
# class RSS_Data_Feeds:
#
#     def __init__(self):
#         pass
#
#
#     def hacker_news(self):
#
#
#         def parse_rss_feed(url):
#             # Send a request to the URL
#             response = requests.get(url)
#             response.raise_for_status()
#
#             # Parse the XML response
#             root = ET.fromstring(response.content)
#
#             # Prepare a list to hold the formatted entries
#             formatted_entries = []
#
#             # Extract and format entries
#             for item in root.findall('.//item'):
#                 formatted_entry = {
#                     "title": item.find('title').text,
#                     "link": item.find('link').text,
#                     "published": item.find('pubDate').text,
#                     "description": item.find('description').text
#                 }
#                 formatted_entries.append(formatted_entry)
#
#             return formatted_entries
#
#         # URL of the RSS feed
#         feed_url = "https://feeds.feedburner.com/TheHackersNews"
#
#
#         entries = parse_rss_feed(feed_url)      # Parse the feed and get formatted entries
#
#         return entries


import os
import uuid
from flask import g, request, session, current_app

from cbr_website_beta.flask.Flask_Utils import Flask_Utils
from osbot_utils.helpers.trace.Trace_Call import Trace_Call
from osbot_utils.helpers.trace.Trace_Call__Config import Trace_Call__Config
from osbot_utils.utils.Misc import date_time_now

from cbr_website_beta.flask.filters.Current_User import Current_User

log_req_count = 0
log_list_all  = []
app_id        = uuid.uuid4().hex[:6]    # unique id of server or lambda function
app_started   = date_time_now()



def request_tracer():
    if hasattr(g, 'request_tracer'):
        return g.request_tracer

    trace_config                          = Trace_Call__Config()
    trace_config.title                    = "Traced Request calls"
    trace_config.trace_ignore_start_with  = ['osbot_utils.utils.Json'] #werkzeug.local','werkzeug.datastructures', 'typing', 'os']
    trace_config.trace_capture_start_with = ['cbr_website_beta', 'flask']
    trace_call                            = Trace_Call(config=trace_config)

    #request_tracer.trace_capture_all = True
    #request_tracer.trace_capture_start_with.append('osbot')

    return trace_call

#@xray_trace("populate_variable_g")
def populate_variable_g():
    if Flask_Utils.is_static_request():
        return

    app = current_app

    if hasattr(app, 'user_data'):                                       # todo: find a better way to set this value, since this is set via Current_User.client__logged_in
        g.user_data = app.user_data
    if hasattr(g, 'user_data') is False or not g.user_data :                                # only set g.user_data if it hasn't been set already (for example some tests willl set this value)
        g.user_data     = Current_User().user_data_from_s3() or {}

    g.user_groups   = g.user_data.get('cognito:groups') or ''
    g.user_name     = g.user_data.get('username')       or ''

    g.trace_call = request_tracer()
    g.data_loaded = 'OK'
    g.execution_env = os.environ.get('EXECUTION_ENV', 'LOCAL')

    g.log_list      = []
    g.log_list_all  = log_list_all
    g.log_req_count = 0
    g.app_id        = app_id
    g.app_started   = app_started
    g.host          = request.headers.get('Host'                       ) or ''
    g.latitude      = request.headers.get('Cloudfront-Viewer-Latitude' ) or ''
    g.longitude     = request.headers.get('Cloudfront-Viewer-Longitude') or ''
    g.ip_address    = request.headers.get('X-Forwarded-For'            ) or ''
    g.aws_trace_id  = request.headers.get('X-Amzn-Trace-Id'            ) or ''



from flask import render_template

from cbr_website_beta.data.odin.Analysis__Chat_Threads import Analysis__Chat_Threads
from cbr_website_beta.flask.filters.Current_User import g_user_data_current_username


HTML_TITLE__USER_CHAT_THREAD = 'User Chat Thread'

class Views__Chat_Threads:

    def __init__(self):
        self.analysis__chat_threads = Analysis__Chat_Threads()

    def chat_thread(self,thread_id):
        chat_threads  = Analysis__Chat_Threads(max_answer_size=0)
        thread_data   = chat_threads.create_consolidated_thread(thread_id)
        kwargs        = dict(thread_id=thread_id, thread_data=thread_data)
        render_kwargs =   { "template_name_or_list" : "user/chat/user-thread-view.html" ,
                            "title"                 : HTML_TITLE__USER_CHAT_THREAD      ,
                            **kwargs                                                    }
        return render_template(**render_kwargs)

    def chat_threads(self):
        user = g_user_data_current_username()
        if user:
            threads_data = self.analysis__chat_threads.threads_data__by_user(user)
            return render_template('user/chat/user-threads.html', threads_data=threads_data)



from functools import cache

from cbr_website_beta.aws.s3.DB_Odin_Data import DB_Odin_Data
from osbot_utils.utils.Dev import pprint


class Analysis__Users_Data:

    def __init__(self):
        self.db_odin_data = DB_Odin_Data()

    def analysis__users_data(self):
        data = self.users_metadatas__data()
        users_data = []
        def add_field(target, source, field_name, source_field_1, source_field_2):
            target[field_name] = source.get(source_field_1) or source.get(source_field_2) or ''


        for user_metadata in data.values():
            user_data = { 'user_id': user_metadata.get('user_id') }
            add_field(user_data, user_metadata, 'first_name'  , 'First name'          , ''                      )
            add_field(user_data, user_metadata, 'last_name'   , 'Last name'           , ''                      )
            add_field(user_data, user_metadata, 'role'        , 'Role'                , 'title'                 )
            add_field(user_data, user_metadata, 'organisation', 'Organisation'        , 'organisation'          )
            add_field(user_data, user_metadata, 'linkedin'    , 'Linkedin'            , 'linkedin'              )
            add_field(user_data, user_metadata, 'country'     , 'Country'             , 'country'               )
            add_field(user_data, user_metadata, 'sector'      , 'Sector'              , 'sector'                )
            add_field(user_data, user_metadata, 'org_size'    , 'Size of organisation', 'size of organisation'  )
            users_data.append(user_data)

            if 'Additional suggested prompts for Athena, your AI advisor' in user_metadata:
                del user_metadata['Additional suggested prompts for Athena, your AI advisor']
            if 'cognito_data' in user_metadata:
                del user_metadata['cognito_data']
        return users_data

    @cache                                      # cache data so that we only make one call to S3 during the lifecycle of this instance
    def users_metadatas(self):
        return self.db_odin_data.users_metadatas()

    def users_metadatas__data(self):
        return self.users_metadatas().get('data')

    def users_metadatas__metadata(self):
        return self.users_metadatas().get('metadata')


from cbr_website_beta.users.CBR_User import CBR_User
from cbr_website_beta.users.CBR_Users_Config import CBR_Users_Config
from osbot_utils.utils.Misc import random_text


class CBR_Users(CBR_Users_Config):

    def __init__(self):
        super().__init__()
        #self.db_users     = DB_Users()
        #self.cognito      = Cognito_IDP()
        self.user_pool_id = 'eu-west-2_22vkoIJRJ'


    # def cognito_user_pool(self):
    #     return self.cognito.user_pool(self.user_pool_id)
    #
    # def cognito_users(self, limit=10):
    #     return self.cognito.users(self.user_pool_id,limit=limit)
    #
    # def cognito_user(self, user_name):
    #     return self.cognito.user(self.user_pool_id, user_name=user_name)

    def user(self, user_id):
        return CBR_User(user_id)

    def users_names(self):
        return self.s3_folder_list(self.s3_folder_users_data())

    def random_user(self):
        user_id = random_text('cbr_random_user').lower()
        return self.user(user_id=user_id)






import traceback

from flask import render_template, Flask

from cbr_website_beta._cbr_shared.dynamo_db.DyDB__CBR_Logging import DyDB__CBR_Logging
from cbr_website_beta.apps.CBR_Config import cbr_config
from cbr_website_beta.flask.decorators.allow_annonymous      import allow_anonymous



def register_error_handling(app: Flask):

    dydb_cbr_logging = DyDB__CBR_Logging()

    @app.errorhandler(Exception)
    def internal_server_error(exception):
        if cbr_config.dev__capture_exceptions():
            dydb_cbr_logging.log_exception(exception)
            return render_template('home/page-500.html', error=str(exception)), 500
        else:
            raise exception

    @app.route('/raise_exception')
    @allow_anonymous
    def cause_error():
        # This will cause an error and trigger the error handler
        return 1 / 0


from cbr_website_beta.aws.s3.DB_Users import DB_Users
from cbr_website_beta.aws.s3.DB_Session import DB_Session
from cbr_website_beta.aws.s3.S3_DB_Base import S3_DB_Base
from osbot_utils.utils.Dev import pprint

class DB_Sessions(S3_DB_Base):

    def db_session(self, user_id):
        return DB_Session(user_id)

    def db_sessions_ids(self):
        return self.s3_folder_list(folder=self.s3_folder_users_sessions())

    def sessions(self):
        for session_id in self.db_sessions_ids():
            yield self.db_session(session_id)

    def sessions__all_data(self, max=None):
        all_data = {}
        for session in self.sessions():
            all_data[session.session_id] = session.session_data()
            if max and len(all_data) >= max:
                break
        #return [session.data() for session in self.sessions()]
        return all_data


from osbot_aws.AWS_Config import AWS_Config
from osbot_aws.apis.S3 import S3
from osbot_aws.apis.Session import Session
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.decorators.methods.remove_return_value import remove_return_value
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Files import file_extension

S3_DEFAULT_FILE_CONTENT_TYPE = 'binary/octet-stream'
S3_HTML_PAGE_CONTENT_TYPE    = 'text/html; charset=utf-8'
S3_FILES_CONTENT_TYPES       = { '.js'  : 'application/javascript; charset=utf-8',
                                 '.jpg' : 'image/jpeg'                           ,
                                 '.jpeg': 'image/jpeg'                           ,
                                 '.png' : 'image/png'                            ,
                                 '.txt' : 'text/plain; charset=utf-8'            ,
                                 '.pdf' : 'application/pdf'                      ,
                                 '.html': 'text/html; charset=utf-8'             ,
                                 '.css' : 'text/css; charset=utf-8'              ,
                                 '.svg' : 'image/svg+xml'                        ,
                                 '.gif' : 'image/gif'                            ,
                                 '.webp': 'image/webp'                           ,
                                 '.json': 'application/json; charset=utf-8'      ,
                                 '.xml' : 'application/xml; charset=utf-8'       ,
                                 '.zip' : 'application/zip'                      ,
                                 '.mp3' : 'audio/mpeg'                           ,
                                 '.mp4' : 'video/mp4'                            ,
                                 '.avi' : 'video/x-msvideo'                      ,
                                 '.mov' : 'video/quicktime'                      }

# todo: add this to the official OSBot_AWS S3 class
class S3(S3):

    @cache_on_self
    def client(self):
        # todo: understand the impact of having to add the endpoint_url below
        #       we needed this because (some, not all, which is really weird) buckets
        #       were throwing this error when creating the pre-signed URL
        #       <Code>SignatureDoesNotMatch</Code>
        #           <Message>The request signature we calculated does not match the signature you provided. Check your key and signing method.</Message>
        #       this is the comment that gave the solution https://github.com/boto/boto3/issues/1149#issuecomment-793737086
        import boto3
        aws_config   = AWS_Config()
        region_name  = aws_config.region_name()
        endpoint_url = f'https://s3.{region_name}.amazonaws.com'
        return boto3.client('s3', region_name=region_name,  endpoint_url=endpoint_url)

    def create_pre_signed_url(self, bucket_name, object_name, operation='get_object', expiration=3600):     # 3600 is one hour (60 seconds * 60 minutes)
        response = self.client().generate_presigned_url(operation, Params={'Bucket': bucket_name, 'Key': object_name}, ExpiresIn=expiration)
        return response
    def file_content_type(self, bucket, key):
        return self.file_details(bucket, key).get('ContentType')

    def file_content_type_update(self, bucket, key, metadata, content_type):
        file_copy_kwargs = dict(bucket_source      = bucket      ,
                                key_source         = key         ,
                                bucket_destination = bucket      ,
                                key_destination    = key         ,
                                metadata           = metadata    ,      # in S3 we can't update the content type with also updating the metadata
                                content_type       = content_type)      # i guess this is related to the fact that the content_type is stored as a metadata item
        return self.file_copy(**file_copy_kwargs)

    @remove_return_value('ResponseMetadata')
    def file_copy(self, bucket_source, key_source, bucket_destination, key_destination, metadata=None, content_type=None):
        kwargs_file_copy = dict(CopySource = {'Bucket': bucket_source, 'Key': key_source},
                                Bucket     = bucket_destination                          ,
                                Key        = key_destination                             )
        if metadata is not None:
            kwargs_file_copy['Metadata'         ] = metadata
            kwargs_file_copy['MetadataDirective'] = 'REPLACE'
        if content_type is not None:
            kwargs_file_copy['ContentType'] = content_type
        return self.client().copy_object(**kwargs_file_copy)

    def file_metadata(self, bucket, key):
        return super().file_details(bucket, key).get('Metadata')

    def file_metadata_update(self, bucket, key, metadata):
        file_copy_kwargs = dict(bucket_source      = bucket  ,
                                key_source         = key     ,
                                bucket_destination = bucket  ,
                                key_destination    = key     ,
                                metadata           = metadata)
        return self.file_copy(**file_copy_kwargs)

    # add support for changing the content type

    def file_upload_to_key(self, file, bucket, key, set_content_type=True):
        extra_args   = None
        if set_content_type:
            extension = file_extension(file)
            if extension:
                content_type = S3_FILES_CONTENT_TYPES.get(extension)
                if content_type:
                    extra_args = {'ContentType': content_type}
        upload_args = [file, bucket, key, extra_args]
        self.client().upload_file(*upload_args)                                # upload file
        return True                                                             # return true (if succeeded)


from flask import render_template

from cbr_website_beta.apps.odin.S3_Backend_Analysis import S3_Backend_Analysis
from osbot_utils.testing.Duration import Duration


HTML_TITLE_ODIN_ACTIONS = 'Odin Actions'

class Views__Dev__Odin:

    def __init__(self):
        self.s3_backend_analysis = S3_Backend_Analysis()

    def api_update_s3_data(self, update_sessions=True, update_users=True):
        with Duration(print_result=False) as duration_update_db_sessions:
            if update_sessions:
                self.s3_backend_analysis.s3_update_db_sessions_status()
        with Duration(print_result=False) as duration_update_users:
            if update_users:
                self.s3_backend_analysis.s3_update_db_users_metadata()
        result = {'duration_update_sessions': round(duration_update_db_sessions.seconds(),2),
                  'duration_update_users'   : round(duration_update_users      .seconds(),2)}
        return result

    def odin_actions(self):
        return render_template(**self.odin_actions__render_config())

    def odin_actions__render_config(self):
        return { "template_name_or_list" : "dev/odin_actions.html" ,
                 "title"                 : HTML_TITLE_ODIN_ACTIONS }


# from functools import wraps
# from cbr_website_beta.flask.middleware.xray_middleware import current_segment_print
# from cbr_website_beta.utils.Web_Utils import Web_Utils
#
# def xray_trace(name):
#     if Web_Utils.running_in_aws() is False:                 # only enable this decorator when running in AEWS
#         return lambda f: f
#     def decorator(f):
#         @wraps(f)
#         def wrapper(*args, **kwargs):
#             subsegment = xray_recorder.begin_subsegment(name)
#             if subsegment:
#                 subsegment.put_http_meta ('url'       , name    )       # this is a hack to add the name to the box that shows in the
#                 subsegment.put_http_meta ('method'    , ''      )       #     xray UI (makes is easier to find the function)
#                 subsegment.put_annotation('trace_name', name    )
#                 #subsegment.put_metadata  ('operation' , name    )
#             try:
#                 result = f(*args, **kwargs)
#             except Exception as e:
#                 subsegment.add_exception(e)
#                 raise
#             finally:
#                 xray_recorder.end_subsegment()
#             return result
#         return wrapper
#     return decorator
#
# from aws_xray_sdk.core import xray_recorder
#
# class XRay_Trace:
#     def __init__(self, name='XRay_Trace', source='pytest', print_segment=False):
#         self.name          = name
#         self.source        = source
#         self.print_segment = print_segment
#
#     def __enter__(self):
#         #if self.skip_pytest_check:
#         self.subsegment = xray_recorder.begin_subsegment(self.name)
#         if self.subsegment:
#             self.subsegment.put_http_meta('url'        , self.name)  # to show name in the x-ray UI
#             self.subsegment.put_http_meta('method'     , ''       )      # placeholder, can be left empty
#             self.subsegment.put_annotation('source'    , self.source)
#             #self.subsegment.put_annotation('annotation', self.name)
#
#     def __exit__(self, exc_type, exc_value, traceback):
#         if self.subsegment:
#             if exc_type is not None:
#                 self.subsegment.add_exception(exc_value, traceback)
#             xray_recorder.end_subsegment()
#         if self.print_segment:
#             current_segment_print()


import os
from functools import cache

from flask import Flask
from importlib import import_module
from flask_minify import Minify

from cbr_website_beta.flask.Flask_Utils                 import FLASK_APP__STATIC_PATH, FLASK_APP__STATIC_URL
from cbr_website_beta.flask.filters.Athena_Html_Content import Athena_Html_Content
from cbr_website_beta.flask.filters.Current_User        import Current_User
from cbr_website_beta.flask.filters.Misc_Filters        import Misc_Filters
from cbr_website_beta.flask.filters.Obj_Data            import Obj_Data
from cbr_website_beta.flask.filters.Pretty_Json         import Pretty_Json
from cbr_website_beta.flask.register_error_handling     import register_error_handling
from cbr_website_beta.flask.register_hooks              import register_hooks
from cbr_website_beta.flask.register_logging            import register_logging
from cbr_website_beta.flask.register_logging_in_g       import register_logging_in_g
from cbr_website_beta.flask.register_middlewares        import register_middlewares
from cbr_website_beta.flask.register_processors         import register_processors
from cbr_website_beta import static
from osbot_utils.decorators.methods.cache_on_self       import cache_on_self

def register_blueprints(app):
    blueprints = ['api', 'llms', 'root' ,'dev', 'home', 'minerva', 'user']
    for module_name in blueprints:
        full_module_name = f'cbr_website_beta.apps.{module_name}.{module_name}_routes'
        module = import_module(full_module_name)
        app.register_blueprint(module.blueprint)

def flask_app_kwargs():
    return dict(static_folder   = FLASK_APP__STATIC_PATH ,
                static_url_path = FLASK_APP__STATIC_URL  )

def create_flask_app():
    app_kwargs = flask_app_kwargs()
    app = Flask(__name__, **app_kwargs)
    return app

def create_app(config):                         # todo: refactor to class
    app = create_flask_app()
    register_error_handling(app)
    register_middlewares   (app)
    app.config.from_object (config)
    register_blueprints    (app)
    register_processors    (app)
    register_hooks         (app)
    register_logging       (app)
    register_logging_in_g  (app)


    app.template_folder = '../apps/templates'
    return app






class Flask_Site:

    def __init__(self):
        pass

    @cache
    def app(self):
        app_config = self.app_config()
        app        = create_app(app_config)
        self.setup(app)
        if not self.debug():
            Minify(app=app, html=True, js=False, cssless=False)
        return app

    def app_config(self):
        from cbr_website_beta.apps.config import config_dict
        return config_dict[self.get_config_mode().capitalize()]


    def debug(self):
        return (os.getenv('DEBUG', 'False') == 'True')

    def get_config_mode(self):
        return 'Debug' if self.debug() else 'Production'

    def setup(self, app):
        self.register_filters(app)

    def register_filters(self, app):
        Athena_Html_Content (app)
        Current_User        (app)
        Pretty_Json         (app)
        Obj_Data            (app)
        Misc_Filters        (app)



import types

from cbr_website_beta.apps.dev.odin.Odin__Panels__Chat_Threads      import Odin__Panels__Chat_Threads
from cbr_website_beta.apps.dev.odin.Odin__Panels__Debug             import Odin__Panels__Debug
from cbr_website_beta.apps.dev.odin.Odin__Panels__Logs              import Odin__Panels__Logs
from cbr_website_beta.apps.dev.odin.Odin__Panels__Real_Time         import Odin__Panels__Real_Time
from cbr_website_beta.apps.dev.odin.Odin__Panels__Session_Management import Odin__Panels__Session_Management


# todo: add documentation and threat model for this class, since it can potentiall be quite an dangerous one
#       since it allows to call methods from classes that are in the target_classes dictionary
#       this danger is mitigated by the fact that the target_classes and target methods are all hardcoded
# todo: add explict RBAC checks and demands for the methods that are being called

class Render_Panels:

    def __init__(self):
        self.target_classes = self.get_target_classes()

    def get_target_classes(self):
        target_classes = { 'odin_debug'             : Odin__Panels__Debug             () ,
                           'odin_session_management': Odin__Panels__Session_Management() ,
                           #'odin_dashboard_widgets' : Odin__Panels__Dashboard_Widgets () ,
                           'odin_chat_threads'      : Odin__Panels__Chat_Threads      () ,
                           'odin_logs'              : Odin__Panels__Logs              () ,
                           'odin_real_time'         : Odin__Panels__Real_Time         () }
        return target_classes

    def resolve_target_method(self, class_name, method_name):
        target_class = self.target_classes.get(class_name)
        if target_class is None:
            return {'status': 'error', 'data': f'no target class found for: {class_name}'}

        if hasattr(target_class, 'exposed_methods') is False:
            return {'status': 'error', 'data': f'no exposed_methods found for: {class_name}'}
        exposed_methods = target_class.exposed_methods()
        if method_name not in exposed_methods:
            return {'status': 'error', 'data': f'no method found for: {method_name} in {class_name}'}
        target_method = exposed_methods.get(method_name)
        return  {'status': 'ok', 'target_method': target_method}


    def render_panel(self, class_name, method_name, **kwargs):
        resolved_target_method = self.resolve_target_method(class_name, method_name)
        if resolved_target_method.get('status') == 'error':
            return resolved_target_method

        target_method: types.MethodType | None = resolved_target_method.get('target_method')

        try:
            return_value = target_method( **kwargs)
        except Exception as error:
            return {'status': 'error', 'data': f'Error: {error}'}
        if type(return_value) is bytes:
            return return_value
        return {'status': 'ok', 'return_value': return_value}



from flask import render_template
from cbr_website_beta.apps.minerva import blueprint

EXPECTED_ROUTES__MINERVA = [ '/minerva', '/minerva/aws-costs']

# @blueprint.app_errorhandler(Exception)           # UndefinedError
# def handle_undefined_error(e):
#     error_details = {
#         "error_message": str(e),
#         "stack_trace": traceback.format_exc(),
#         # Add more context-specific data here if needed
#     }
#     return render_template('error_page.html'), 500

@blueprint.route('')
def minerva_root():
    return render_template('minerva/index.html')

# todo: fix this route when minerva is fixed (also add caching layer to minerva since there
#       is no point of making more than one call per 8 hours)
@blueprint.route('/aws-costs')
def aws_costs():
    # chart_data = Chart_Data__AWS__Minerva().aws_cost_explorer()
    # chart_labels = chart_data.get('all_days')
    # chart_series = chart_data.get('all_costs')
    # all_services = chart_data.get('all_services')
    chart_labels = {}
    chart_series = {}
    all_services = {}
    return render_template('minerva/aws-costs.html',
                           chart_labels=chart_labels,
                           chart_series=chart_series,
                           all_services=all_services)



# from osbot_aws.apis.Cloud_Watch import Cloud_Watch
# from osbot_utils.utils.Dev import pprint
#
#
# class Cloud_Watch_Metrics:
#     def __init__(self):
#         self.cloud_watch = Cloud_Watch()
#
#     def dashboard_widget(self, dashboard_name, widget_name):
#         widgets = self.dashboard_widgets(dashboard_name=dashboard_name)
#         return widgets.get(widget_name, {})
#
#     def dashboard_widgets(self, dashboard_name):
#         dashboard = self.cloud_watch.dashboard(dashboard_name=dashboard_name)
#         widgets = {}
#         for widget in dashboard.get('widgets'):
#             properties = widget.get('properties')
#             title      = properties.get('title')
#             widgets[title] = properties
#         return widgets
#
#     def dashboards(self):
#         return self.cloud_watch.dashboards()
#
#     def widget_screenshot(self, dashboard_name, widget_name, save_to_disk=False, start="-PT2H", end="P0D"):
#         widget_data = self.dashboard_widget(dashboard_name, widget_name)
#         widget_data['start' ] = start
#         widget_data['end'   ] = end
#         widget_data['period'] = 300
#         return self.cloud_watch.metric_widget_image(widget_data, save_to_disk=save_to_disk)
#
#     # def metric_list(self, namespace):
#     #     return self.cloud_watch.metric_list(namespace=namespace)
#     #
#     # def metric_statistics(self, namespace, metric_name, period, start_time, end_time, statistics):
#     #     return self.cloud_watch.metric_statistics(namespace=namespace, metric_name=metric_name, period=period, start_time=start_time, end_time=end_time, statistics=statistics)
#



import os

from osbot_utils.utils.Http import current_host_online


class Web_Utils:

    @staticmethod
    def currently_online():
        return current_host_online()

    # @staticmethod
    # def current_execution_env():
    #     return os.environ.get('EXECUTION_ENV', 'LOCAL')

    @staticmethod
    def running_in_aws():
        return 'AWS_EXECUTION_ENV' in os.environ

    @staticmethod
    def in_aws_code_build():
        return 'CODEBUILD_BUILD_ID' in os.environ


import io
from flask                                                      import request, jsonify, render_template, send_file

from cbr_website_beta.apps.CBR_Config import cbr_config
from cbr_website_beta.apps.dev                                  import blueprint
from cbr_website_beta.apps.dev.Render_Panels                    import Render_Panels
from cbr_website_beta.apps.dev.odin.Odin__Panels__Chat_Threads  import Odin__Panels__Chat_Threads
from cbr_website_beta.apps.dev.odin.Views__Dev__Odin            import Views__Dev__Odin
from cbr_website_beta.apps.dev.view_data                        import useful_links, for_danny, prompt_uis, user_management
from cbr_website_beta.apps.dev.views.Debug_Views                import Debug_Views
from cbr_website_beta.apps.dev.views.Real_Time_Views            import Real_Time_Views
from cbr_website_beta.apps.dev.views.Render_View                import Render_View
from cbr_website_beta.apps.dev.views.Users_Views                import Users_Views
from cbr_website_beta.bots.Athena_Rest_API                      import Athena_Rest_API
from cbr_website_beta.flask.decorators.allow_annonymous         import admin_only
from cbr_website_beta.utils.Web_Utils                           import Web_Utils

EXPECTED_METHODS_DEV = [ 'dev_blueprint.all_chat_threads'              ,
                         'dev_blueprint.api_odin_update_s3_data'       ,
                         'dev_blueprint.dashboard_widget'              ,
                         'dev_blueprint.dashboard_widgets'             ,
                         'dev_blueprint.dev_client_details'            ,
                         'dev_blueprint.dev_current_sessions'          ,
                         'dev_blueprint.dev_current_user'              ,
                         'dev_blueprint.dev_current_users'             ,
                         'dev_blueprint.dev_debug_views'               ,
                         'dev_blueprint.dev_logs_streams'              ,
                         'dev_blueprint.dev_real_time_code_editor'     ,
                         'dev_blueprint.dev_render_panel'              ,
                         'dev_blueprint.dev_request_details'           ,
                         'dev_blueprint.dev_root'                      ,
                         'dev_blueprint.dev_view'                      ,
                         'dev_blueprint.logs_chat_threads'             ,
                         'dev_blueprint.logs_ip_address'               ,
                         'dev_blueprint.logs_odin'                     ,
                         'dev_blueprint.logs_user_sessions'            ,
                         'dev_blueprint.logs_web'                      ,
                         'dev_blueprint.odin_chat'                     ,
                         'dev_blueprint.view_thread_id'                ]

EXPECTED_ROUTES__DEV = [ '/dev/'                                       ,
                         '/dev/all-chat-threads'                       ,
                         '/dev/api/odin/update-s3-data'                ,
                         '/dev/chat/<thread_id>'                       ,
                         '/dev/client-details'                         ,
                         '/dev/current-sessions'                       ,
                         '/dev/current-user/<user_id>'                 ,
                         '/dev/current-users'                          ,
                         '/dev/dashboard/widget/<widget_name>/<start>' ,
                         '/dev/dashboard/widgets'                      ,
                         '/dev/debug-views'                            ,
                         '/dev/dev-panel'                              ,
                         '/dev/logs-chat-threads'                      ,
                         '/dev/logs-web'                               ,
                         '/dev/logs-odin'                              ,
                         '/dev/logs-sessions'                          ,
                         '/dev/logs/ip-address'                        ,
                         '/dev/odin/odin-chat'                         ,
                         '/dev/real-time/streams'                      ,
                         '/dev/real-time/code-editor'                  ,
                         '/dev/request-details'                        ,
                         '/dev/view/<class_name>/<method_name>'        ]

EXPECTED_DEV_MENU = {   "All Chat Threads" : "/dev/all-chat-threads"     ,
                        "Client Details"   : "/dev/client-details"       ,
                        'Code Editor'      : '/dev/real-time/code-editor',
                        "Current Sessions" : "/dev/current-sessions"     ,
                        "Current Users"    : "/dev/current-users"        ,
                        "Dev"              : "/dev/"                     ,
                        'Debug Views'      : '/dev/debug-views'          ,
                        'Ip Address'       : '/dev/logs/ip-address'      ,
                        'Logs Chat Threads': '/dev/logs-chat-threads'    ,
                        'Logs Odin'        : '/dev/logs-odin'            ,
                        'Logs Sessions'    : '/dev/logs-sessions'        ,
                        'Logs Web'         : '/dev/logs-web'             ,
                        'Odin Chat'        : '/dev/odin/odin-chat'       ,
                        "Request Details"  : "/dev/request-details"      ,
                        'Widgets'          : '/dev/dashboard/widgets'    ,
                        'Streams'          : '/dev/real-time/streams'    }

@blueprint.route('/', strict_slashes=False)
@admin_only
def dev_root():
    bot_athena = Athena_Rest_API()
    #git_repo_status = bot_athena.git_repo_status()
    return render_template('dev/index.html',
                           useful_links    = useful_links    ,
                           for_danny       = for_danny       ,
                           prompt_uis      = prompt_uis      ,
                           user_management = user_management )
                           #github_repos=github_repos,

                           #git_repo_status= git_repo_status)

@blueprint.route('/api/odin/update-s3-data')
@admin_only
def api_odin_update_s3_data():
    result = Views__Dev__Odin().api_update_s3_data()
    return jsonify(result)

@blueprint.route('/all-chat-threads')
@admin_only
def all_chat_threads():
    class_name  = 'odin_chat_threads'
    method_name = 'all_chat_threads'
    return Render_View().render_view(class_name, method_name)

@blueprint.route('/chat/<thread_id>')
@admin_only
def view_thread_id(thread_id):
    #class_name    = 'odin_chat_threads'
    #method_name   = 'chat_thread'
    #method_kwargs = dict(thread_id=thread_id)
    return Odin__Panels__Chat_Threads().chat_thread(thread_id)
    #return Render_View().render_view(class_name, method_name, **method_kwargs)


@blueprint.route('/client-details')
@admin_only
def dev_client_details():               return Debug_Views().client_details()

@blueprint.route('/dashboard/widgets')
@admin_only
def dashboard_widgets():
    class_name = 'odin_dashboard_widgets'
    method_name = 'dashboard_metrics'
    return Render_View().render_view(class_name, method_name)

@blueprint.route('/dashboard/widget/<widget_name>/<start>')
@admin_only
def dashboard_widget(widget_name, start):
    render_panels = Render_Panels()
    class_name    = 'odin_dashboard_widgets'
    method_name   = 'widget_screenshot'
    method_kwargs = dict(widget_name = widget_name,
                         start      = f"-PT{start}H"             ,
                         end        = "P0D"               )
    render_kwargs = dict(class_name  = class_name  ,method_name = method_name, **method_kwargs)
    image_bytes   = render_panels.render_panel(**render_kwargs)

    result_io = io.BytesIO(image_bytes)
    result_io.seek(0)  # Go to the beginning of the file-like object

    # Return the file-like object as a PNG image
    return send_file(result_io, mimetype='image/png', as_attachment=False)


@blueprint.route('/logs-chat-threads')
@admin_only
def logs_chat_threads():
    current_env   = cbr_config.env()
    hours         = request.args.get('hours'    , default=1          , type=int)
    env           = request.args.get('partition', default=current_env, type=str)
    class_name    = 'odin_logs'
    method_name   = 'cbr_chat_threads'
    method_kwargs = dict(hours=hours, env=env)
    return Render_View().render_view(class_name, method_name, **method_kwargs)

@blueprint.route('/logs-web')
@admin_only
def logs_web():
    current_env   = cbr_config
    hours         = request.args.get('hours'    , default=1          , type=int)
    env           = request.args.get('partition', default=current_env, type=str)
    class_name    = 'odin_logs'
    method_name   = 'cbr_requests'
    method_kwargs = dict(hours=hours, env=env)
    return Render_View().render_view(class_name, method_name, **method_kwargs)

@blueprint.route('/logs-odin')
@admin_only
def logs_odin():
    current_env   = cbr_config.env()
    hours         = request.args.get('hours'    , default=1          , type=int)
    env           = request.args.get('partition', default=current_env, type=str)
    class_name    = 'odin_logs'
    method_name   = 'cbr_logging'
    method_kwargs = dict(hours=hours, env=env)
    return Render_View().render_view(class_name, method_name, **method_kwargs)

@blueprint.route('/logs-sessions')
@admin_only
def logs_user_sessions():
    current_env   = cbr_config.env()
    hours         = request.args.get('hours'    , default=1          , type=int)
    env           = request.args.get('partition', default=current_env, type=str)
    class_name    = 'odin_logs'
    method_name   = 'cbr_user_sessions'
    method_kwargs = dict(hours=hours, env=env)
    return Render_View().render_view(class_name, method_name, **method_kwargs)

@blueprint.route('/logs/ip-address')
@admin_only
def logs_ip_address():
    current_env   = cbr_config.env()
    hours         = request.args.get('hours'     , default=5          , type=int)
    env           = request.args.get('env'       , default=current_env, type=str)
    ip_address    = request.args.get('ip_address', default='NA'       , type=str)
    class_name    = 'odin_logs'
    method_name   = 'ip_address'
    method_kwargs = dict(hours=hours, env=env, ip_address=ip_address)
    return Render_View().render_view(class_name, method_name, **method_kwargs)

@blueprint.route('/odin/odin-chat')
@admin_only
def odin_chat():
    class_name  = 'odin_real_time'
    method_name = 'chat_prompt'
    return Render_View().render_view(class_name, method_name)


@blueprint.route('/real-time/streams')
@admin_only
def dev_logs_streams():
    class_name  = 'odin_real_time'
    method_name = 'log_streams'
    return Render_View().render_view(class_name, method_name)

@blueprint.route('/real-time/code-editor')
@admin_only
def dev_real_time_code_editor():
    class_name  = 'odin_real_time'
    method_name = 'code_editor'
    return Render_View().render_view(class_name, method_name)

@blueprint.route('/request-details')
@admin_only
def dev_request_details():              return Debug_Views().request_details()

# @blueprint.route('/request-logs')
# @admin_only
# def dev_request_logs():
#     return Logs_Views().request_logs()

@blueprint.route('/current-sessions')
@admin_only
def dev_current_sessions():
    class_name = 'odin_session_management'
    method_name = 'current_sessions'
    return Render_View().render_view(class_name, method_name)

    #return Sessions_Views().current_sessions()

@blueprint.route('/current-user/<user_id>')
@admin_only
def dev_current_user(user_id:str=None):
    return Users_Views().current_user(user_id)

@blueprint.route('/current-users')
@admin_only
def dev_current_users():
    class_name  = 'odin_session_management'                                      # todo: find a better to map these Render_View methods
    method_name = 'current_users'
    return Render_View().render_view(class_name, method_name)
    #return Users_Views().current_users()


@blueprint.route('/dev-panel', methods=['POST'])
@admin_only
def dev_render_panel():                                 # todo refactor this from this class into the Render_Panels class
    if request.is_json:
        try:
            data          = request.get_json()
            class_name    = data.get('class_name')
            method_name   = data.get('method_name')
            method_kwargs = data.get('method_kwargs', {})

            if not class_name or not method_name:
                return jsonify({'status': 'error', 'message': 'class_name and method_name are required'}), 400

            result = Render_Panels().render_panel(class_name, method_name, **method_kwargs)
            return jsonify(result)
        except Exception as error:
            return jsonify({'status': 'error', 'message': f'Error: {error}'}), 500
    else:
        return jsonify({'status': 'error', 'message': 'Request must be JSON'}), 400

@blueprint.route('/view/<class_name>/<method_name>')
@admin_only
def dev_view(class_name, method_name):
    return Render_View().render_view(class_name, method_name)


# @blueprint.route('/odin-actions')
# @admin_only
# def dev_odin_actions():
#     return Views__Dev__Odin().odin_actions()

# @blueprint.route('/tree-view')
# @admin_only
# def dev_tree_view():
#     return view__tree_view()

# @blueprint.route('/s3-explorer')
# @admin_only
# def dev_s3_explorer():
#     class_name  = 'odin_s3_explorer'
#     method_name = 's3_browser'
#     return Render_View().render_view(class_name, method_name)

@blueprint.route('/debug-views')
@admin_only
def dev_debug_views():
    class_name    = 'odin_debug'
    method_name   = 'debug_views'
    method_kwargs = dict()
    return Render_View().debug_views(class_name, method_name, ** method_kwargs)
    #return view__tree_view()



import sys

from flask import Flask

app : Flask
try:
    from cbr_website_beta.flask.Flask_Site          import Flask_Site
    from osbot_utils.utils.Files                    import path_combine

    path_to_add = path_combine(__file__, '..')        # this is needed to that the apps folder can be resolved
    sys.path.append(path_to_add)

    flask_site = Flask_Site()
    app = flask_site.app()

    if __name__ == "__main__":
        app.run()

except Exception as error:
    print()
    print("***********************************")
    print("********* ERROR in run.py *********")
    print(error)
    print("***********************************")
    print()


import sys
from datetime import datetime
from os import environ

from flask import request, g

from cbr_website_beta._cbr_shared.dynamo_db.DyDB__CBR_Requests import DyDB__CBR_Requests
from cbr_website_beta._cbr_shared.schemas.CBR_Request import CBR_Request
from cbr_website_beta.flask.Flask_Utils import Flask_Utils
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Misc import date_time_now, list_set

#LOG_STATIC_FILES = True
def env_vars_to_log():
    env_vars = {'lambda': {"memory_size"        : environ.get('AWS_LAMBDA_FUNCTION_MEMORY_SIZE' ,''),
                           "function_name"      : environ.get('AWS_LAMBDA_FUNCTION_NAME'        ,''),
                           "function_version"   : environ.get('AWS_LAMBDA_FUNCTION_VERSION'     ,''),
                           "aws_region"         : environ.get('AWS_REGION'                      ,''),
                           "execution_env"      : environ.get('EXECUTION_ENV'                   ,''),
                           "amzn_trace_id"      : environ.get('_X_AMZN_TRACE_ID'                ,''),
                           "runtime"            : environ.get('AWS_EXECUTION_ENV'               ,'')}}
    return env_vars

#@xray_trace("register_logging")
def register_logging(app):
    if 'pytest' in sys.modules:
        return


    dydb_cbr_requests = DyDB__CBR_Requests()

    #todo: move this logic to the post_request so that we can capture the elapsed time
    @app.before_request
    #@xray_trace("before_request - pre_request_logging")
    def pre_request_logging():
        if Flask_Utils.is_static_request():
            return
        try:
            request._start_time = datetime.utcnow()

            if '/llms-ui' in request.path :
                return
            headers = {key: value for key, value in request.headers.items()}
            country = headers.get('Cloudfront-Viewer-Country-Name')             # todo refactor to a country and city method
            if not country:                                                     # if 'Cloudfront-Viewer-Country-Name' did not had a value
                country = headers.get('Cloudfront-Viewer-Country', 'NA')        #    try Cloudfront-Viewer-Country

        #@app.after_request
        #def post_request_logging(response):                                    # todo: refactor into helper logging class and methods

            cbr_request            = CBR_Request()
            cbr_request.user       = g.user_name         or 'NA'
            cbr_request.user_role  = f'{g.user_groups}'  or 'NA'                # todo , fix naming convention
            cbr_request.ip_address = g.ip_address        or 'NA'
            cbr_request.path       = request.path        or 'NA'
            cbr_request.date       = date_time_now(date_time_format='%Y-%m-%d')
            cbr_request.level      = "DEBUG"
            cbr_request.method     = request.method      or 'NA'
            cbr_request.referer    = request.referrer    or 'NA'
            cbr_request.headers    = headers
            cbr_request.environ    = env_vars_to_log()
            cbr_request.city       = headers.get('Cloudfront-Viewer-City'     , 'NA')
            cbr_request.country    = country
            #cbr_request.env_list   = list_set(environ)
            #cbr_request.req_id     = ???
            # cbr_request.session_id     = ???
            cbr_request.source     = 'CBR Website'


            #cbr_request.status_code= ???   #todo add this to the response section
            document = cbr_request.json()
            #pprint(document)
            dydb_cbr_requests.add_document(document)
        except Exception as e:
            pprint(e)               # todo add custom error handler which will send the errors to DyDB

        # timestamp    = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')  # ISO 8601 format
        # request_data = {"env"            : g.execution_env                      ,
        #                 "user_name"      : g.user_name                          ,
        #                 "user_groups"    : g.user_groups                        ,
        #                 #"remote_addr"   : request.remote_addr                  ,
        #                 "ip_address"     : g.ip_address                         ,
        #                 #"timestamp"     : datetime.utcnow().strftime("%d/%b/%Y %H:%M:%S"),
        #                 #"time"           : datetime.utcnow().strftime("%H:%M:%S"),
        #                 #"method"        : request.method                        ,
        #                 "path"           : request.path                          ,
        #                 "latitude"       : g.latitude                            ,
        #                 "longitude"      : g.longitude                           ,
        #                 #"scheme"        : request.scheme.upper()                ,
        #                 #"protocol"      : request.environ.get("SERVER_PROTOCOL").split("/")[-1],
        #                 }


        # todo: figure out why this takes so long to run in AWS (between 20ms to 70ms)


        #with Duration(f'cloud_watch_logs.send_log'):
        #cloud_watch_logs.send_log(log_msg)
        #with Duration(f'DyDB__Timeseries.add_document'):
        #dydb_timeseries  .add_document(log_msg, partition=g.execution_env)
        #dydb_web_requests.add_document(log_msg, partition=g.execution_env)

    # @app.after_request
    # def post_request_logging(response):
    #     if hasattr(request, "_start_time"):
    #         request_latency = datetime.utcnow() - request._start_time
    #         request_latency_seconds = request_latency.total_seconds()
    #     else:
    #         request_latency_seconds = 0
    #
    #     response_data = { "timestamp"      : datetime.utcnow().strftime("%d/%b/%Y %H:%M:%S"),
    #                       "status_code"    : response.status_code                           ,
    #                       "latency_seconds": request_latency_seconds                        }
    #     log_msg = {"source": "flask_logging" ,
    #                "type"  : 'response_data' ,
    #                "data"  : response_data  }
    #
    #     #logger.info(log_msg)
    #     cloud_watch_logs.send_log(log_msg)
    #
    #     return response


# from osbot_utils.utils.Dev import pprint
# from osbot_utils.utils.Json import json_dumps, json_loads
# from osbot_aws.apis.Logs import Logs
#
#
# class Cloud_Watch_Logs:
#
#     def __init__(self):
#         self.log_name          = 'cbr_website_beta.sep.2023'
#         self.stream_name__10_m = 'real_time_stream__10_minutes'
#         self.stream_name__1_h  = 'real_time_stream__1_hour'
#         self.logs__10_m        = Logs(group_name=self.log_name, stream_name=self.stream_name__10_m)
#         self.logs__1_h         = Logs(group_name=self.log_name, stream_name=self.stream_name__1_h )
#         try:
#             self.logs__10_m.create()
#             self.logs__1_h .create()
#         except Exception as error:
#             pprint(f'**** Error in Cloud_Watch_Logs: {error}' )
#
#     def send_log(self, log_data):
#         log_data['via'] = 'Cloud_Watch_Logs'
#         message = json_dumps(log_data)
#
#         result__10m = self.logs__10_m.event_add(message)    # todo: fix stream , since at the moment it is not deleting data after 10m
#         #result__1h  = self.logs__1_h .event_add(message)   # for now we only send to the 10m stream
#         return dict(result__10m=result__10m)#, result__1h=result__1h)
#
#     def get_logs(self, logs, limit=10000, hours=24):              # 10000 is the maximum allowed by AWS (or 1M)
#         messages = logs.messages(limit=limit, hours=hours)
#
#         log_data = []
#         for message in messages:
#             log_data.append(json_loads(message))
#         return log_data
#
#     def get_logs__10_m(self, limit=10000, hours=1):
#         return self.get_logs(self.logs__10_m, limit, hours=hours)
#
#     def get_logs__1_h(self, count=10000):
#         return self.get_logs(self.logs__1_h, count)


#from cbr_website_beta.aws.dynamodb.DyDB__Chat_Threads import DyDB__Chat_Threads
from cbr_website_beta.aws.s3.S3_DB_Base import S3_DB_Base
from osbot_utils.utils.Misc import timestamp_utc_now, timestamp_to_str_date, timestamp_to_str_time

FILE_NAME_CURRENT_SESSIONS = 'current-sessions.json'
FILE_NAME_USERS_METADATA   = 'users-metadata.json'

class DB_Odin_Data(S3_DB_Base):

    def chat_threads(self):
        return []
        #return DyDB__Chat_Threads().documents()

    def current_sessions(self):
        return self.current_sessions__from_dydb()

    def current_sessions__from_dydb(self):
        from cbr_website_beta.aws.dynamodb.DyDB__Sessions import DyDB__Sessions         # todo: fix circular dependency with the DB_Odin_Data and DyDB__Sessions (the S3 methods below need to be refactored in to separate class )
        dydb_sessions = DyDB__Sessions()
        raw_sessions_data = dydb_sessions.sessions()
        sessions_data = {}
        if raw_sessions_data:
            for raw_session_data in raw_sessions_data:                      # todo: refactor this into a helper method
                session_id = raw_session_data.get('session_id')            #       we need to do this at the moment the data is supposed to be a dict
                sessions_data[session_id] = raw_session_data                #       and indexed by the session_id

        return {'metadata': self.file_metadata('Odin', source='DyDB'),
                'data'    : sessions_data }

    def current_sessions__from_s3(self):
        return self.s3_file_data(self.s3_key__current_sessions())

    def current_sessions__save_to_s3(self, data, saved_by='Odin'):
        file_data = { 'metadata' : self.file_metadata(saved_by=saved_by, source='S3'),
                      'data'     : data                                 }
        return self.s3_save_data(file_data, self.s3_key__current_sessions())

    def file_metadata(self, saved_by, source):
        timestamp      = timestamp_utc_now()
        timestamp_data = timestamp_to_str_date(timestamp)
        timestamp_time = timestamp_to_str_time(timestamp)
        return dict (timestamp      = timestamp     ,
                     timestamp_date = timestamp_data,
                     timestamp_time = timestamp_time,
                     saved_by       = saved_by      ,
                     source         = source       )

    def s3_key__current_sessions(self):
        return f'{self.s3_folder_odin_data()}/{FILE_NAME_CURRENT_SESSIONS}'

    def s3_key__users_metadatas(self):
        return f'{self.s3_folder_odin_data()}/{FILE_NAME_USERS_METADATA}'


    def users_metadatas(self):
        return self.s3_file_data(self.s3_key__users_metadatas())

    def users_metadatas__save(self, users_metatada, saved_by='NA', runtime='NA'):
        file_data = { 'metadata' : self.file_metadata(saved_by, runtime),
                      'data'     : users_metatada                       }
        return self.s3_save_data(file_data, self.s3_key__users_metadatas())


from datetime import datetime
from functools import cache

from cbr_website_beta.aws.s3.DB_Odin_Data import DB_Odin_Data
from cbr_website_beta.aws.s3.DB_Sessions import DB_Sessions

from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Misc import timestamp_to_str_date, timestamp_to_str_time


class Analysis__Users_Sessions:

    def __init__(self):
        self.db_odin_data = DB_Odin_Data()

    def analysis__sessions_by_user(self):
        data = self.sessions__data()
        consolidated_view = {}

        for session in data.values():
            username  = session['data']['username']
            timestamp = session['timestamp']
            role      = session['data'].get('cognito:groups')
            if type(role) is list:
                role = role[0]
            else:
                role = 'normal user'

            if username not in consolidated_view:
                consolidated_view[username] = {'number_of_sessions': 1, 'latest_timestamp': timestamp, 'username': username }
                consolidated_view[username]['session_date'] = timestamp_to_str_date(timestamp)
                consolidated_view[username]['session_time'] = timestamp_to_str_time(timestamp)
                consolidated_view[username]['user_role'   ] = role
                consolidated_view[username]['session_id'  ] = session['session_id']
            else:
                consolidated_view[username]['number_of_sessions'] += 1
                if timestamp > consolidated_view[username]['latest_timestamp']:
                    consolidated_view[username]['latest_timestamp' ] = timestamp
                    consolidated_view[username]['session_date'     ] = timestamp_to_str_date(timestamp)
                    consolidated_view[username]['session_time'     ] = timestamp_to_str_time(timestamp)
                    consolidated_view[username]['session_id'       ] = session['session_id']


        # for user in consolidated_view.values():
        #     del user['latest_timestamp']

        return list(consolidated_view.values())

    @cache                                                      # cache data so that we only make one call to S3 during the lifecycle of this instance
    def sessions(self):
        return self.db_odin_data.current_sessions()

    def sessions__data(self):
        return self.sessions().get('data')

    def sessions__metadata(self):
        return self.sessions().get('metadata')


    def table_data__current_sessions(self, data):
        table_data = []

        for session_id, session_data in data.items():
            data         = session_data.get('data', {})
            user_groups  = data.get('cognito:groups'   )
            username     = data.get('username'         )
            user_id      = data.get('sub'              )
            source       = session_data.get('source'   )
            timestamp    = session_data.get('timestamp')
            row = dict( username    = username                           ,
                        date        = timestamp_to_str_date(timestamp)   ,
                        time        = timestamp_to_str_time(timestamp)   ,
                        user_groups = user_groups                        ,
                        session_id  = session_id                         ,
                        source      = source                             ,
                        user_id     = user_id                            )

            table_data.append(row)

        return table_data

    def view_data__current_sessions(self):
        data        = self.sessions__data()
        metadata    = self.sessions__metadata()
        table_data  = self.table_data__current_sessions(data)
        return { 'table_data' : table_data, 'metadata' : metadata }



from cbr_website_beta.aws.s3.DB_Users import DB_Users


class Temp_DB_User:

    def __init__(self, prefix='pytest__temp_db_user'):
        self.db_users = DB_Users()
        self.db_users.s3().dont_use_threads()
        self.db_user  = self.db_users.random_db_user(prefix=prefix)

    def __enter__(self):
        assert self.db_user.create() is True
        return self.db_user

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert self.db_user.delete() is True



import os

import jwt

from osbot_aws.apis.Cognito_IDP import Cognito_IDP
from osbot_utils.decorators.methods.cache_on_self import cache_on_self


class Cognito_Login:
    def __init__(self):
        self.cognito = Cognito_IDP()

    @cache_on_self
    def client_id(self):
        return os.environ.get("COGNITO_CLIENT_ID")

    def region(self):
        return os.environ.get("AWS_DEFAULT_REGION")

    def project(self):
        return os.environ.get("COGNITO_PROJECT")

    def user_pool_id(self):
        return os.environ.get("COGNITO_USER_POOL_ID")

    def login(self, username, password):
        login_result = {}
        result               = self.cognito.auth_initiate(self.client_id(), username, password)
        if result.get('error'):
            login_result['status'] = 'error'
            login_result['error' ] = str(result.get('error'))
        else:
            auth_result          = result.get('auth_result')
            raw_access_token     = auth_result.get('AccessToken')
            decoded_access_token = jwt.decode(raw_access_token, options={"verify_signature": False})
            login_result['status'      ] = 'ok'
            login_result['access_token'] = decoded_access_token
        return login_result


from cbr_website_beta import static
from cbr_website_beta.aws.s3.S3 import S3
from cbr_website_beta.utils.Site_Utils import Site_Utils
from osbot_aws.AWS_Config import AWS_Config

from cbr_website_beta.utils.Version import version
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Files import files_list, path_combine

BUCKET__STATIC_ASSETS  = '{account_id}-cbr'
S3_FOLDER__CLOUD_FRONT = 'cbr_website_static'
FOLDER_NAME__DIST      = 'dist'
S3_FOLDER__DIST        = f'{S3_FOLDER__CLOUD_FRONT}/{FOLDER_NAME__DIST}'


class Create_Deploy_Assets_In_S3(Kwargs_To_Self):
    aws_config : AWS_Config
    version    : str

    def __init__(self):
        super().__init__()
        self.version = version

    @cache_on_self
    def s3(self):
        return S3()

    def map_local_files_to_s3_keys(self):
        path_static_files = path_combine(static.path, FOLDER_NAME__DIST)

        js_files          = files_list(path_static_files, '*.js')
        files_to_upload   = {}
        for local_file_path in js_files:
            virtual_file_path = local_file_path.replace(path_static_files, '')
            s3_key            = self.s3_key_for_uploaded_file(virtual_file_path)
            files_to_upload[s3_key] = local_file_path
        return files_to_upload

    def s3_key_for_uploaded_file(self, virtual_path):
        s3_key = f"{S3_FOLDER__DIST}/{self.version}{virtual_path}"
        return s3_key

    def target_bucket(self):
        return BUCKET__STATIC_ASSETS.format(account_id=self.aws_config.account_id())

    def upload_files_to_s3(self):
        print()
        s3_bucket         = self.target_bucket()
        s3_keys_for_files = self.map_local_files_to_s3_keys()
        for s3_key, local_file_path in s3_keys_for_files.items():
            print(f'uploading: {s3_key}')
            self.s3().file_upload_to_key(local_file_path, s3_bucket, s3_key)
        return f'Uploaded {len(s3_keys_for_files)} files to s3'




import cbr_website_beta
from osbot_utils.base_classes.Type_Safe import Type_Safe
from osbot_utils.utils.Files            import file_contents, path_combine

class Version(Type_Safe):

    FILE_NAME_VERSION = 'version'

    def path_code_root(self):
        return cbr_website_beta.path

    def path_version_file(self):
        return path_combine(self.path_code_root(), self.FILE_NAME_VERSION)

    def value(self):
        value = file_contents(self.path_version_file()) or ""
        return value.strip()

version = Version().value()


from functools import cache

from markdown import Markdown

from cbr_website_beta.aws.s3.DB_Odin_Data import DB_Odin_Data
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Lists import list_index_by, list_group_by
from osbot_utils.utils.Misc import list_set, timestamp_to_str_date, timestamp_to_str_time

DEFAULT_MAX_ANSWER_SIZE = 100

class Analysis__Chat_Threads(Kwargs_To_Self):
    db_odin_data    : DB_Odin_Data
    max_answer_size : int = DEFAULT_MAX_ANSWER_SIZE
    markdown_answer = True


    def all_consolidated_threads(self):
        results = {}
        for thread_id in self.threads_ids():
            results[thread_id] = self.create_consolidated_thread(thread_id)
        return results

    # def threads_ids_per_user(self):
    #     results = {}
    #     for thread_id, threads_data in self.threads_data__by_id().items():
    #         user = threads_data[0].get('user')
    #         results[user] = results.get(user, []) + [thread_id]
    #     return []


    def create_consolidated_thread(self, thread_id):
        by_thread_id = self.threads_data__by_id()
        thread_data = by_thread_id.get(thread_id)
        thread_data = sorted(thread_data, key=lambda x: x['chat_thread']['index'], reverse=False)

        self.remove_system_prompts(thread_data)
        self.add_index_to_root   (thread_data)

        combined_data = []

        # Use a range with a step of 2 since data is in pairs
        for i in range(0, len(thread_data), 2):
            prompt_entry       = thread_data[i]                                               # Assuming the prompt is always followed by its corresponding answer
            answer_entry       = thread_data[i + 1] if i + 1 < len(thread_data) else {}
            index              = prompt_entry.get('index')                                           # Extract index from the prompt entry
            user               = prompt_entry.get('user')
            user_prompt        = self.get_user_prompt(prompt_entry)
            answer             = self.get_answer     (answer_entry)
            answer_size        = self.get_answer_size(answer_entry)
            prompt_timestamp   = int(prompt_entry.get('chat_thread', {}).get('timestamp',0))
            prompt_date        = timestamp_to_str_date(prompt_timestamp, date_format='%d %b')
            prompt_time        = timestamp_to_str_time(prompt_timestamp)
            answer_timestamp   = int(answer_entry.get('chat_thread', {}).get('timestamp', 0))
            prompt_with_images = len(prompt_entry.get('chat_thread', {}).get('data', {}).get('images', [])) > 0
            prompt_model       = prompt_entry.get('chat_thread', {}).get('data', {}).get('model' , '')

            timespan = (answer_timestamp - prompt_timestamp) / 1000
            combined_entry = {  'index'             : index              ,
                                'prompt'            : user_prompt        ,
                                'prompt_date'       : prompt_date        ,
                                'prompt_time'       : prompt_time        ,
                                'prompt_timestamp'  : prompt_timestamp   ,
                                'prompt_with_images': prompt_with_images ,
                                'prompt_model'      : prompt_model       ,
                                'answer'            : answer             ,
                                'answer_size'       : answer_size        ,
                                'answer_timestamp'  : answer_timestamp   ,
                                'user'              : user               ,
                                'timespan'          : timespan           }

            combined_data.append(combined_entry)
        return combined_data

    def get_user_prompt(self, entry):
        chat_thread = entry.get('chat_thread', {})
        data        = chat_thread.get('data', {})
        return data.get('user_prompt', 'NA')

    def get_answer(self, entry):
        max_answer_size = self.max_answer_size
        answer_raw_data = self.get_answer_raw_data(entry)
        if max_answer_size:
            if len(answer_raw_data) > max_answer_size:
                return answer_raw_data[:max_answer_size] + " ...."
            return answer_raw_data[:max_answer_size]
        if self.markdown_answer:
            return self.markdown().convert(source=answer_raw_data)                   # todo: security: review for XSS injection via Markdown payloads
        return answer_raw_data

    def get_answer_raw_data(self, entry):
        chat_thread     = entry.get('chat_thread', {})
        answer_raw_data = chat_thread.get('answer', '')
        return answer_raw_data

    def get_answer_size(self, entry):
        return len(self.get_answer_raw_data(entry))


    # def get_histories(self, entry):
    #     chat_thread = entry.get('chat_thread', {})
    #     data       = chat_thread.get('data', {})
    #     return data.get('histories', '')


    def add_index_to_root(self, threads_data):
        for item in threads_data:
            chat_thread = item.get('chat_thread',{})
            index       = int(chat_thread.get('index'))
            item['index'] = index

    @cache_on_self
    def markdown(self):
        return Markdown(extensions=['tables'])

    def remove_system_prompts(self, threads_data):
        for item in threads_data:
            self.remove_system_prompt(item)

    def remove_system_prompt(self, item):
        chat_thread = item.get('chat_thread',{})
        data        = chat_thread.get('data', {})
        if 'system_prompts' in data:
            del data['system_prompts']

    @cache_on_self
    def threads_data(self):
        return self.db_odin_data.chat_threads()

    def threads_data__by_id(self):
        return list_group_by(self.threads_data(), 'thread_id')

    def threads_data__by_user(self, user):
        return self.user__consolidated_threads(user)

    def threads_ids(self):
        return list_set(self.threads_data__by_id())

    def user__consolidated_threads(self, user):
        return self.users__consolidated_threads().get(user, {})

    def users__consolidated_threads(self):
        results = {}
        for thread_id, consolidated_thread in self.all_consolidated_threads().items():
            user = consolidated_thread[0].get('user')
            results[user] = results.get(user, {})
            results[user][thread_id] = consolidated_thread
        return results





from flask import g

from cbr_website_beta.apps.llms.llms_routes import current_user_data
from cbr_website_beta.flask.decorators.allow_annonymous import allow_anonymous
from cbr_website_beta.apps.api import blueprint
from cbr_website_beta.flask.filters.Current_User import Current_User, g_user_data

EXPECTED_ROUTES__API = ['/api/user-data']

@blueprint.route('/user-data')
@allow_anonymous
def user_data():
    return { "g.user_data": g_user_data()      ,
             "cognito"   : current_user_data() }



# from datetime import datetime, timedelta
#
# from osbot_utils.utils.Json import json_loads
#
# from osbot_utils.decorators.methods.cache_on_self import cache_on_self
#
# from osbot_aws.apis.Logs import Logs
# from typing import cast
#
# CBR_WEBSITE_LOG_GROUP_NAME = 'cbr_website_beta.flask.Flask_Site'
#
# class Cloud_Watch_Logs__Read:
#
#     @cache_on_self
#     def logs(self) -> Logs:
#         return Logs()
#
#     def log_groups(self):
#         logs = cast(Logs, self.logs())
#         return logs.groups_names()
#
#     def group_name(self):
#         return CBR_WEBSITE_LOG_GROUP_NAME
#
#     def generate_s3_keys(self, start_time, end_time, interval_minutes):
#         s3_keys = []
#         current_time = start_time - timedelta(minutes=start_time.minute % interval_minutes)
#         while current_time < end_time:
#             folder = current_time.strftime("%d/%m/%Y")
#             file_name = f"logs_{current_time.hour:02}_{current_time.minute:02}.json"
#             s3_keys.append(f"/{folder}/{file_name}")
#             current_time += timedelta(minutes=interval_minutes)
#         return s3_keys
#
#
#
#     def fetch_logs_from_cloudwatch(self, start_time, end_time):
#         cloudwatch = self.logs().client()
#         response = cloudwatch.filter_log_events(
#             logGroupName=self.group_name(),
#             startTime=start_time,
#             endTime=end_time
#         )
#
#         logs = []
#         for event in response["events"]:
#             logs.append(json_loads(event["message"]))
#
#         return logs


from cbr_website_beta.apps.dev.odin.Odin__Panels__Chat_Threads import Odin__Panels__Chat_Threads
from cbr_website_beta.apps.user.user_profile import user_profile
from cbr_website_beta.apps.user.views.Views__Chat_Threads import Views__Chat_Threads
from cbr_website_beta.aws.cognito.Cognito_Auth_Flow import Cognito_Auth_Flow
#from cbr_website_beta.aws.dynamodb.DyDB__Chat_Threads import DyDB__Chat_Threads
from cbr_website_beta.flask.decorators.allow_annonymous import allow_anonymous
from cbr_website_beta.apps.user import blueprint
from flask import redirect, request, make_response, render_template, jsonify, g

from cbr_website_beta.flask.filters.Current_User import DEFAULT_ADMIN_GROUPS
from cbr_website_beta.flask.utils.current_server import current_server
from osbot_utils.utils.Misc import random_guid, random_uuid

LOCAL_DEV_SERVER     = 'http://localhost:5000/'
COGNITO_PROJECT      = 'the-cbr-beta'
COGNITO_REGION       = 'eu-west-2'
COGNITO_CLIENT_ID    = '5ij6l5kdho4umoks5rjfh9cbid'
COGNITO_SIGN_IN      = f'https://{COGNITO_PROJECT}.auth.{COGNITO_REGION}.amazoncognito.com/login?client_id={COGNITO_CLIENT_ID}&response_type=code&scope=email+openid+phone&'
COGNITO_SIGN_OUT     = f'https://{COGNITO_PROJECT}.auth.{COGNITO_REGION}.amazoncognito.com/logout?client_id={COGNITO_CLIENT_ID}&response_type=code&scope=email+openid+phone&'
EXPECTED_USER__HOME  = [ '/admin/impersonate_user/<user_token>' ,
                         '/admin/restore_admin_user'            ,
                         '/login'                               ,
                         '/logout'                              ,
                         '/sign-in'                             ,
                         '/sign-out'                            ,
                         '/unauthorized'                        ,
                         '/user/chat-threads'                   ,
                         '/user/chat/<thread_id>',
                         '/user/profile'                        ]


@blueprint.route('/login')
@allow_anonymous
def login():
    url = COGNITO_SIGN_IN + f"redirect_uri={current_server()}sign-in"
    return redirect(url)


@blueprint.route('/sign-in')
@allow_anonymous
def sign_in():

    sign_in_code      = request.args.get('code')                    # todo: refactor this logic out of this method

    cognito_auth_flow = Cognito_Auth_Flow()
    cookie_data = cognito_auth_flow.create_cbr_token_cookie_from_cognito_code(sign_in_code=sign_in_code)
    if not cookie_data:
        return {'error': 'could not authenticate user, please contact the Support team'}   #todo create a better error page

    cbr_token       = cookie_data.get('cookie_value')
    user_info       = cookie_data.get('user_info' , {})
    role            = user_info.get('cognito:groups')
    render_kwargs   = {"template_name_or_list": "/home/accounts/logging_in.html",
                     "session_id" : cbr_token                                        }
    response_html   = render_template(**render_kwargs)
    #response_html     = render_template('home/accounts/logging_in.html')
    response        = make_response(response_html)


    if role == DEFAULT_ADMIN_GROUPS:                                            # if the user is admin then use both values for cookie (which is mainly used for making UI decisions,i.e. this CANNOT be a security control)
        cbr_token = f"{cbr_token}|{cbr_token}"                                  # todo: figure out why seeting up two cookies here didn't work, initial version of this code was seeting an CBR_ADMIN_TOKEN, but that was not working in prod
        response.set_cookie('CBR_TOKEN', cbr_token)
    else:
        response.set_cookie('CBR_TOKEN', cbr_token)
    return response




@blueprint.route('/unauthorized')
@allow_anonymous
def unauthorized():
    return render_template('home/accounts/unauthorized.html')

@blueprint.route('/sign-out')
def sign_out():
    redirect_to = redirect(current_server())
    response    = make_response(redirect_to)
    response.set_cookie('CBR_TOKEN'      , '', expires=0)
    #response.set_cookie('CBR_ADMIN_TOKEN', '', expires=0)                  # todo: figure out why this is not working, in prod, only the first cookie is beeing reset ok
    return response

# todo: see if we need this (since at the moment this is not wired)
@blueprint.route('/logout')
@allow_anonymous
def logout():
    url = COGNITO_SIGN_OUT + f"logout_uri={current_server()}sign-out"

    return redirect(url)

@blueprint.route('/user/profile', methods=['GET', 'POST'])
@allow_anonymous
def profile():
    return user_profile()


@blueprint.route('/admin/impersonate_user/<user_token>', methods=['GET'])            # todo: change to POST method
def admin__impersonate_user(user_token):
    redirect_to = redirect(current_server())
    response = make_response(redirect_to)
    from flask import g
    if g.user_data:                                                                 # todo: refactor all this login into an impersonation class
        if g.user_data.get('admin_token'):
            admin_token = g.user_data.get('admin_token')
            impersonated_cookie_value = f'{user_token}|{admin_token}'
            response.set_cookie('CBR_TOKEN', impersonated_cookie_value)
    return response

@blueprint.route('/admin/restore_admin_user', methods=['GET'])                      # todo: change to POST method
def admin__restore_admin_user():
    redirect_to = redirect(current_server())
    response = make_response(redirect_to)
    from flask import g
    if g.user_data:                                                                  # todo: refactor all this login into an impersonation class
        if g.user_data.get('admin_token'):
            admin_token = g.user_data.get('admin_token')
            restored_cookie_value = f'{admin_token}|{admin_token}'
            response.set_cookie('CBR_TOKEN', restored_cookie_value)
    return response

@blueprint.route('/user/chat-threads')
def user__chat_threads():
    return Views__Chat_Threads().chat_threads()


@blueprint.route('/user/chat/<thread_id>')
def user_chat_thread(thread_id):
    return Views__Chat_Threads().chat_thread(thread_id)
    #return Odin__Panels__Chat_Threads().chat_thread(thread_id)
    #return Render_View().render_view(class_name, method_name, **method_kwargs)


# @blueprint.route('/user/save-chat', methods=['POST'])
# def user_save_chat():
#     return jsonify({'status': 'warning', 'message': 'temporarily disabled'})
    # if not g.user_name:
    #     return jsonify({'status': 'error', 'message': 'there was no user available'}), 400
    #
    # if request.is_json:
    #     try:
    #         user_name = g.user_name
    #         post_data   = request.get_json()
    #
    #         thread_id   = post_data.get('thread_id') or random_uuid()
    #         index       = post_data.get('index')
    #         thread_data = post_data.get('thread_data')
    #         data_type   = post_data.get('data_type')
    #         dydb_chat_threads = DyDB__Chat_Threads()
    #         result = dydb_chat_threads.add_chat_thread(user_name, thread_id, index, data_type, thread_data)
    #         return jsonify({'status': 'ok', 'result': result}), 200
    #     except Exception as error:
    #         return jsonify({'status': 'error', 'message': f'Error: {error}'}), 500
    # else:
    #     return jsonify({'status': 'error', 'message': 'Request must be JSON'}), 400

